"""Definition of the interface class for the gymnasium environment definition.

The gym environment is a class that incorporates the interface requirements set
out by the OpenAI gymnasium library for Reinforcement Learning environment.
This standard is very commonly used in python based Reinforcement Learning
applications and most python libraries supplying Reinforcement Learning agents
can use this interface.

This interface was not incorporated into the ReLeSO.parser_environment class
due to incompatibility between the pydantic and gym functionality. It is also
advantageous to separate the two classes since now it is possible to make
multiple instances of the environment accessible for multi environment training
and separate validation environment.

An instance of this class is created by the
ReLeSO.parser_environment.Environment.get_gym_environment function. Each call
of this function creates a new gym environment.

Note:
    Currently it is not possible to run xns solvers concurrently out of the box
    since currently all created environments use the same run location. This
    has the result that the output files of the xns solver are also created in
    the same folder for all environments and therefore overwrite each other. A
    fix for this problem is currently worked on.
"""

from typing import Any, List, Union

import gymnasium

from releso.util.types import StepReturnType


class GymEnvironment(gymnasium.Env):
    """Environment interface class for the gym environment definition.

    This class is a placeholder class which complies with the OpenAI gym
    Interface. The real functionality is infused into the class after creation
    by substituting the functions from the
    SbSOvRL.parser_environment.Environment class.
    """

    metadata = {"render.modes": ["mesh"]}

    def __init__(self, action_space, observation_space) -> None:
        """Constructor for gym environment object.

        Many functions are dummy implementations and are overwritten by the
        correct functions at run time.

        Args:
            action_space (_type_): _description_
            observation_space (_type_): _description_
        """
        super().__init__()
        self.action_space = action_space
        self.observation_space = observation_space

    def step(self, action: Union[int, List[float]]) -> StepReturnType:
        """Dummy function definition for gym interface."""

    def reset(self) -> Any:
        """Dummy function definition for gym interface."""

    def render(self, mode: str = "mesh") -> None:
        """Dummy function definition for gym interface."""

    def close(self) -> None:
        """Dummy function definition for gym interface."""



"""SPOR definition Read Carefully.

Solver Postprocessing Observation Reward (SPOR) represents all steps which

follow after completing Geometry adaption, where the geometry is
deformed via the actions.

This part of the ReLeSO environment definition is a list of steps the
environment will run through in series. Each step can be configured to use the
ReLeSO communication protocol.

A more in depth documentation of the SPOR concept is given here
:ref:`SPOR Communication Interface <sporcominterface>`
"""

import importlib
import json
import os
import pathlib
import shutil
import sys
import traceback
from ast import literal_eval
from timeit import default_timer as timer
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import uuid4

import numpy as np
from pydantic import UUID4, Field, conint, validator
from pydantic.fields import PrivateAttr

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.observation import (
    ObservationDefinition,
    ObservationDefinitionMulti,
)
from releso.util.logger import VerbosityLevel, set_up_logger
from releso.util.reward_helpers import spor_com_parse_arguments
from releso.util.types import ObservationType, RewardType, StepReturnType
from releso.util.util_funcs import JSONEncoder, call_commandline, join_infos


class MultiProcessor(BaseModel):
    """Definition of the utilized multiprocessor. Default MPIEXEC.

    The MultiProcessor gives access to command prefixes which can enable
    parallelization. Currently only tested for MPI.

    Note:
        The cluster uses PATH variables to correctly call mpi. IF you want to
        use MPI multiprocessing on the cluster please use
        :py:class:`ReLeSO.spor.MPIClusterMultiProcessor`.
    """

    #: command prefix which is needed to parallelize the give task
    command: str = "mpiexec -n"
    #: Max core count to use for this parallelization option. This variable
    #: is not shared between tasks or environments so the user has to split
    #: these up before hand.
    max_core_count: conint(ge=1)
    #: Whether or not to always use the multiprocessing option. This is
    #: helpful if multiprocessing command prefix is needed even if core count
    #: is 1. Defaults to *False*.
    always_use: bool = False

    def get_command(self, core_count: int) -> str:
        """Return the string that represents the call for multiprocessing.

        Args:
            core_count (int): Number of wanted cores.

        Returns:
            str: string representing the command line call prefix
        """
        resulting_core_count = max(1, min(core_count, self.max_core_count))
        return (
            f"{self.command} {resulting_core_count}"
            if resulting_core_count > 1 or self.always_use
            else ""
        )


class MPIClusterMultiProcessor(MultiProcessor):
    """Multi-processor extension for the cluster MPI call."""

    #: The location title needs to be set to ``cluster`` so that
    location: Literal["cluster"]
    # : command prefix which is needed to parallelize the give task

    command: str = "$MPIEXEC"
    #: additional flags for the mpi call.
    mpi_flags_variable: str = "$FLAGS_MPI_BATCH"

    def get_command(self, core_count: int) -> str:
        """Return the string that represents the call for multiprocessing.

        Note:
            Since the cluster uses PATH variables for its computation. This
            function will read these in and use them correspondingly. But will
            not change them. This limitation is done so that there is less
            interference during Multi-Environment training.

        Args:
            core_count (int): Number of wanted cores.

        Raises:
            RuntimeError: This error is thrown if the mpi_flags PATH variable
            is not of the shape expected by the function. Expected are 2 parts
            -np [core_count] where -np can be anything and the core_count
            should be a number.

        Returns:
            str: string representing the command line call prefix
        """
        # self.get_logger().warning("Using Cluster mpi")
        # check if environment variable for mpi has the correct core number.
        resulting_core_count = max(1, min(core_count, self.max_core_count))
        if resulting_core_count == 1 and not self.always_use:
            return ""
        environ_mpi_flags: List = str(
            os.environ[self.mpi_flags_variable.replace("$", "")]
        ).split()
        local_mpi_flags = self.mpi_flags_variable
        if len(environ_mpi_flags) == 0:
            core_qualifier = "-n"
            set_core_count = -1
        elif len(environ_mpi_flags) == 2:
            core_qualifier, set_core_count = environ_mpi_flags
        else:
            self.get_logger().error(
                "The environment variable for mpi cluster did not look as "
                "expected. Please use standard MultiProcessor or revise the "
                "code. The variable is as follow "
                f"{environ_mpi_flags, len(environ_mpi_flags)}"
            )
            raise RuntimeError(
                "MPI Cluster environment variable not as expected. "
                "More information in the log."
            )
        if set_core_count != resulting_core_count:
            local_mpi_flags = f"{core_qualifier} {resulting_core_count}"
        self.get_logger().debug(
            f"Using Cluster mpi with command "
            f"{os.environ[self.command.replace('$', '')]} and additional "
            f"flags {local_mpi_flags}"
        )
        return f"{self.command} {local_mpi_flags}"


class SPORObject(BaseModel):
    """Base class SPORObject can not be instantiated.

    Base class for all possible SPOR object classes. Theses objects
    represent the possible steps that can be/are executed after the FFD
    finished.

    The following arguments need to be set for SPOR objects, further
    parameters are use-case specific.
    """

    #: name of the SPOR step. Makes so that it is possible to distinguish
    #: different steps more easily.
    name: str
    #: Whether to stop after this step if this step has thrown an error.
    #: This is important if
    stop_after_error: bool = True
    #: Reward which is returned if any kind of error is thrown during
    #: completing the defined task.
    reward_on_error: float
    #: Reward which is returned if the task completed with out an error. If
    #: this is set this reward will overwrite any other reward this function
    #: will generate. (Implementation specific might differ for child classes
    #: see documentation). If *None* no default reward is returned when task is
    #: finished. Defaults to *None*.
    reward_on_completion: Optional[float] = None
    #: This boolean can disable the running of the task when a reset is
    #: performed.
    run_on_reset: bool = True
    #: How many additional observations does this object yield. Please define
    #: this correctly and not just write in a single number. Be responsible,
    #: the single number functionality is only in here due to backwards
    #: compatibility reasons.
    additional_observations: Union[
        ObservationDefinition,
        ObservationDefinitionMulti,
        List[Union[ObservationDefinition, ObservationDefinitionMulti]],
        None,
    ] = None

    #: Whether or not the first time setup still needs to be done.
    _first_time_setup_not_done: bool = PrivateAttr(True)

    @validator("additional_observations", pre=True)
    @classmethod
    def validate_additional_observations(
        cls, v: str, values: Dict[str, Any]
    ) -> Union[
        ObservationDefinition,
        ObservationDefinitionMulti,
        List[Union[ObservationDefinition, ObservationDefinitionMulti]],
    ]:
        """Validator additional_observations.

        Validates the the additional observations variable. This is a pre
        validation function only used for back compatibility reasons.

        Args:
            v (str): See pydantic
            values (Dict[str, Any]):  See pydantic

        Returns:
            Union[
                ObservationDefinition, ObservationDefinitionMulti,
                List[ObservationDefinition]]: Pre validated values.
        """
        if not isinstance(v, (dict, list)):
            if not int(v) == 0:
                # We might want to deprecate the hidden option to instantiate
                #  additional observations via an int. Building up a default
                #  observation definition.
                # TODO: change name to something more meaningful for example
                # using the function name values["name"]
                v = {
                    "name": f"unnamed_{str(uuid4())}",
                    "value_min": -1,
                    "value_max": 1,
                    "observation_shape": [int(v)],
                    "value_type": "float",
                    "save_location": values["save_location"],
                }

            else:
                v = None

        return v

    def get_observations(
        self,
    ) -> Union[
        ObservationDefinition,
        ObservationDefinitionMulti,
        List[Union[ObservationDefinition, ObservationDefinitionMulti]],
        None,
    ]:
        """Return number of observation the step generates.

        Returns the number of observations which are generated by this
        SPORStep.

        Returns:
            Union[
                ObservationDefinition, ObservationDefinitionMulti,
                List[ObservationDefinition]]: Additional observation
                definitions
        """
        return self.additional_observations

    def get_default_observation(
        self, observations: ObservationType
    ) -> ObservationType:
        """Generate default observations if no were returned or step failed.

        Generates default observations for all additional observations defined
        in this object.

        Args:
            observations (ObservationType): Already defined observations to
            which the additional observations are added to.

        Returns:
            ObservationType: Observations with now the additional observations
            of this object added (values used are the default values)
        """
        if self.additional_observations is None:
            pass  # observation do not need to be changed
        elif isinstance(self.additional_observations, list):
            for item in self.additional_observations:
                observations[item.name] = item.get_default_observation()
        else:
            observations[self.additional_observations.name] = (
                self.additional_observations.get_default_observation()
            )
        return observations

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """Performs the defined step.

        This function is called to complete the defined step of the current
        SPORObject. This functions needs to be overloaded by all child classes.

        Note:
            Please use these default values for the return values: observation
            = None, reward = 0, done = False, info = dict()

        Args:
            step_information (StepReturnType): Previously collected step
            values. Should be up-to-date.
            environment_id (UUID4): Environment ID which is used to distinguish
            different environments which run in parallel.
            validation_id (Optional[int], optional): During validation the
            validation id signifies the validation episode. If none the current
            episode is not a validation episode. Defaults to None.
            core_count (int, optional): Wanted core_count of the task to run.
            Defaults to 1.
            reset (bool, optional): Boolean on whether or not this object is
            called because of a reset. Defaults to False.


        Returns:
            StepReturnType:
                The full compliment of the step information are returned.
                The include observation, reward, done, info.
        """
        raise NotImplementedError


class SPORObjectExecutor(SPORObject):
    """Base definition of a SPORCommandLineObject. Can be instantiated.

    Base class for all possible SPOR object classes which use the command
    line. This class is meant to add some default functionality, so that
    users do not need to add them themselves. These are:

    1. access to the (MPI) Multi-processors via
        :py:class:`ReLeSO.spor.MultiProcessor`
    2. command line command
    3. additional flags
    4. working directory
    5. whether or not to use the
        :ref:`SPOR Communication Interface <sporcominterface>`

    """

    #: Definition of the multi-processor. Defaults to *None*.
    multi_processor: Optional[
        Union[MultiProcessor, MPIClusterMultiProcessor]
    ] = None
    #: whether or not to use the SPOR communication interface.
    #: Defaults to *False*.
    use_communication_interface: bool = False
    # : whether or not to add the step information to the SPOR communications
    #: interface commandline options including the observations, info, done,
    #: reward
    add_step_information: bool = False
    #: Path to the directory in which the program should run, helpful for
    #: programs using relative paths. If `{}` are the first characters of the
    #: path, they will be replaced by the save_location of the current run. If
    #: `{}` are present in the path, the current environment id will be used to
    #: replace the placeholder. This is necessary for multiprocessing.
    working_directory: str

    # communication_interface_variables
    #: If the communication interface is used this UUID will be used to
    #: identify the job/correct worker
    _run_id: UUID4 = PrivateAttr(default=None)

    @validator("working_directory")
    @classmethod
    def validate_working_directory_path(cls, v: str):
        """Validator working_directory.

        Checks if the given path points to a correct directory. If a {} is
        present in the path only the path stub before this substring is
        validated. This functionality can be used for multiprocessing if each
        process needs to be in its own folder.

        Args:
            v (str): Object to validate

        Raises:
            ParserException: If path is not correct, this error is
            thrown.

        Returns:
            str: original path if no validation error occurs.
        """
        # check if placeholder for multiprocessing is available if not use
        # v to validate directory path
        if not v.startswith("{}"):
            path = (
                pathlib.Path(v.split("{}")[0])
                if "{}" in v
                else pathlib.Path(v)
            )
            if not (path.is_dir() and path.exists()):
                raise ParserException(
                    "SPORObjectCommandline",
                    "unknown",
                    f"The work_directory path {v} does not exist or is not a"
                    " valid directory."
                    " If you are using the multiprocessing placeholder {}, make "
                    "sure, that the path before the placeholder is valid and "
                    "exists.",
                )
        return v

    @validator("add_step_information")
    @classmethod
    def validate_add_step_information_only_true_if_also_spor_com_is_true(
        cls, v: str, values: Dict[str, Any]
    ):
        """Validator add_step_information.

        Check that if the validated variable is true the value of the variable
        ``use_communication_interface`` is also true.

        Args:
            v (str): Object to validate
            values (str): Already validated values

        Raises:
            ParserException: If path is not correct, this error is thrown.

        Returns:
            str: value
        """
        if v:
            if (
                "use_communication_interface" in values
                and values["use_communication_interface"]
            ):
                pass
            else:
                raise ParserException(
                    "SPORObjectCommandline",
                    "add_step_information",
                    "Please only set the add_step_information variable to "
                    "True if the variable use_communication_interface is also "
                    "True.",
                )
        return v

    def setup_working_directory(self, environment_id: UUID4):
        """Set up the working directory for the SPORStep.

        Args:
            environment_id (UUID4): Id of the environment. This is the working
                directory.
        """
        if self.working_directory.startswith("{}"):
            self.working_directory = self.working_directory.format(
                self.save_location, environment_id
            )
            pathlib.Path(self.working_directory).expanduser().resolve().mkdir(
                parents=True, exist_ok=True
            )
            self.get_logger().info(
                f"Found placeholder for step with name {self.name}. New "
                f"working directory is {self.working_directory}. Directory was created."
            )
        elif "{}" in self.working_directory:
            self.working_directory = self.working_directory.format(
                environment_id
            )

            # create run folder if necessary
            path = pathlib.Path(self.working_directory).expanduser().resolve()
            if not path.exists():
                path.mkdir(parents=True, exist_ok=True)
                self.get_logger().info(
                    f"Found placeholder for step with name {self.name}. New "
                    f"working directory is {self.working_directory}. "
                    "Directory was created."
                )
            else:
                self.get_logger().info(
                    f"Found placeholder for step with name {self.name}. New "
                    f"working directory is {self.working_directory}. "
                    "Directory already exists, step will use it."
                )
            # TODO this is a XNS specific case
            if self.name == "main_solver":  # pragma: no cover
                shutil.copyfile(path.parent / "xns_multi.in", path / "xns.in")
                self.get_logger().info("Copying file.....")

    def get_multiprocessing_prefix(self, core_count: int) -> str:
        """Add commandline prefix for mpi multiprocessor.

        Function which tries to get the correct multiprocessing command if
        available else an empty string is returned.

        Args:
            core_count (int): Maximal number of cores the multiprocessor can
            use.

        Returns:
            str: command string for the multi processor
        """
        ret_str = ""
        if self.multi_processor:
            ret_str = self.multi_processor.get_command(core_count)
        return ret_str

    def spor_com_interface(
        self,
        reset: bool,
        environment_id: UUID4,
        validation_id: Optional[int],
        step_information: StepReturnType,
    ) -> List[str]:
        """Add spor com interface command line options.

        Generates the additional command line argument options used for the
        spor com interface.

        Args:
            reset (bool): Boolean on whether or not this object is called
                because of a reset.
            validation_id (Optional[int]): During validation the validation id
                signifies the validation episode. If none the current episode
                is not a validation episode.
            step_information (StepReturnType): Previously collected step
                values. Should be up-to-date.
            environment_id (UUID4): Environment ID which is used to distinguish
                different environments which run in parallel.

        Returns:
            List[str]: List of additional command line options.
        """
        # return empty list if the com interface is not used
        if not self.use_communication_interface:
            return []
        # initialize run id if not already existing
        if not self._run_id:
            self._run_id = uuid4()

        # collect all parts of the communication interface
        interface_options: List[str] = ["--run_id", str(self._run_id)]
        interface_options.extend([
            "--base_save_location",
            str(self.save_location),
        ])
        interface_options.extend(["--environment_id", str(environment_id)])
        if reset:
            interface_options.append("--reset")
        if validation_id is not None:
            interface_options.extend([
                "--validation_value",
                str(validation_id),
            ])
        if self.add_step_information:
            json_step_information = {
                "observations": step_information[0],
                "reward": step_information[1],
                "done": step_information[2],
                "info": step_information[3],
            }
            # print(json_step_information)
            interface_options.extend([
                "--json_object",
                "'" + json.dumps(json_step_information, cls=JSONEncoder) + "'",
            ])
        return interface_options

    def spor_com_interface_read(self, output: bytes, step_dict: Dict):
        """Read in return values of the spor com interface.

        Interprets in the printed values from the spor com interface and will
        add them to the observation, reward, done and info return values.

        Args:
            output (bytes): Output of the command line call. To be interpreted.
            step_dict (Dict): Already known values of the step return this dict
            will be updated in place.
        """
        try:
            returned_step_dict = literal_eval(
                output.decode(sys.getdefaultencoding())
            )
        except SyntaxError as err:
            self.get_logger().warning(
                f"An error was thrown while reading in the spor_com_interface "
                f"return values of the step {self.name}. Please check that "
                "only the dictionary holding the correct information is "
                "printed during the execution of the external program.",
                exc_info=True,
            )
            raise SyntaxError from err
        self.spor_com_interface_add(returned_step_dict, step_dict)

    def spor_com_interface_add(
        self, returned_step_dict: Dict[str, Any], step_dict: Dict[str, Any]
    ):
        """Add returned step information of the spor com interface.

        Add the newly returned values to the observation, reward, done and
        info return values.

        Args:
            returned_step_dict (Dict): Newly received values.
            step_dict (Dict): Already known values of the step return this dict
            will be updated in place.
        """
        # add observations
        if self.additional_observations is None:
            pass  # observation do not need to be changed
        elif isinstance(self.additional_observations, list):
            for item in self.additional_observations:
                step_dict["observation"][item.name] = returned_step_dict[
                    "observations"
                ][item.name]
        else:
            step_dict["observation"][self.additional_observations.name] = (
                np.array(returned_step_dict["observations"])
            )

        join_infos(
            step_dict["info"][self.name],
            returned_step_dict["info"],
            self.logger_name,
        )
        if value := step_dict["info"].get(self.name):
            if value := value.get("reset_reason"):
                step_dict["info"]["reset_reason"] = value
        step_dict["done"] = step_dict["done"] or returned_step_dict["done"]
        step_dict["reward"] = returned_step_dict["reward"]

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """Run method for the Execution types of SPORSteps.

        Abstract method.

        Args:
            step_information (StepReturnType): _description_
            environment_id (UUID4): _description_
            validation_id (Optional[int], optional): _description_.
                Defaults to None.
            core_count (int, optional): _description_. Defaults to 1.
            reset (bool, optional): _description_. Defaults to False.

        Raises:
            NotImplementedError: _description_

        Returns:
            StepReturnType: _description_
        """
        raise NotImplementedError


class SPORObjectPythonFunction(SPORObjectExecutor):
    """Base definition of a SPORObjectPythonFunction. Parent class.

    Base class for all possible SPOR python object classes which use
    internalized python function mechanism.
    This class is meant to add some default functionality, so that
    users do not need to add them themselves. These are:

    1. Let an internalized python function be run and save the function data
    2. working directory
    3. whether or not to use the
        :ref:`SPOR Communication Interface <sporcominterface>`

    """

    #: function handler used for python functions which are loaded inside the
    #: program envelop from the outside
    _run_func: Optional[Any] = PrivateAttr(default=None)
    #: logger which is used for the internalized python functions
    _run_logger: Optional[Any] = PrivateAttr(default=None)
    #: persistent function data is not touched by releso
    _func_data: Optional[Any] = PrivateAttr(default=None)

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """This function loads and executes the defined python file.

        The herein defined python file needs to have a function called `main`
        with three parameters.

        1. args: Namespace see SPORCommInterface
        2. logger: logger to be used in the function
        3. func_data: data variable can be used to store persistent data


        Args:
            step_information (StepReturnType):
                Previously collected step values. Should be up-to-date.
            environment_id (UUID4):
                Environment ID which is used to distinguish different
                environments which run in parallel.
            validation_id (Optional[int], optional):
                During validation the validation id signifies the validation
                episode. If none the current episode is not a validation
                episode. Defaults to None.
            core_count (int, optional):
                Can only be one since no multi processing is possible in this
                case.
            reset (bool, optional):
                Boolean on whether or not this object is called because of a
                reset. Defaults to False.

        Returns:
            StepReturnType:
                The full compliment of the step information are returned.
                The include observation, reward, done, info.
        """
        env_logger = self.get_logger()
        step_return: Dict = {
            "observation": step_information[0],
            "reward": step_information[1],
            "done": step_information[2],
            "info": step_information[3],
        }
        # skips the step without doing anything if reset and not run_on_reset
        # add observations if the skipped step has some
        if reset and not self.run_on_reset:
            if self.additional_observations is not None:
                step_return["observation"] = self.get_default_observation(
                    step_return["observation"]
                )
        else:  # executes the step
            args = spor_com_parse_arguments(
                self.spor_com_interface(
                    reset=reset,
                    environment_id=environment_id,
                    validation_id=validation_id,
                    step_information=step_information,
                )
            )
            exit_code = 0
            output = None
            current_dir = os.getcwd()
            try:
                os.chdir(self.working_directory)
                output, self._func_data = self._run_func(
                    args, self._run_logger, self._func_data
                )
            except Exception as err:  # noqa: BLE001
                self.get_logger().warning(
                    f"Could not run the internalized python spor function "
                    f"without error. The following error was thrown {err}."
                    " Please check if this is a user error."
                    f"Traceback is {traceback.format_exc()}."
                )
                exit_code = 404
            os.chdir(current_dir)

            step_return["info"][self.name] = {
                "output": output,
                "exit_code": int(exit_code),
            }
            self.get_logger().debug(
                f"The return step return of the current "
                f"command is: {step_return}"
            )
            if exit_code != 0:  # error thrown during command execution
                env_logger.info(f"SPOR Step {self.name} thrown error.")
                if self.stop_after_error:
                    env_logger.info(
                        "Will stop episode now, due to thrown error."
                    )
                    step_return["done"] = True
                    step_return["info"]["reset_reason"] = (
                        f"ExecutionFailed-{self.name}"
                    )
                step_return["reward"] = self.reward_on_error
                if self.additional_observations is not None:
                    step_return["observation"] = self.get_default_observation(
                        step_return["observation"]
                    )
            else:
                if self.use_communication_interface:
                    if isinstance(output, dict):
                        self.spor_com_interface_add(output, step_return)
                    else:  # this should never be accessed with python funcs
                        raise RuntimeError(
                            "The output of the internalized python function "
                            "is not a dictionary. Please check the function "
                            "and the documentation."
                        )  # pragma: no cover
                        # self.spor_com_interface_read(output, step_return)
                if self.reward_on_completion:
                    if step_return["reward"]:
                        env_logger.warning(
                            f"A reward {step_return['reward']} is already set "
                            f"but a default reward {self.reward_on_completion}"
                            " on completion is also set. The default reward "
                            "will overwrite the already set reward. Please "
                            "check if this is the wanted behavior, if not set "
                            "the reward on completion to None."
                        )
                    step_return["reward"] = self.reward_on_completion

        return (
            step_return["observation"],
            step_return["reward"],
            step_return["done"],
            step_return["info"],
        )


class SPORObjectInternalPythonFunction(SPORObjectPythonFunction):
    """This class will load SPOR functions provided by this package.

    Currently available functions are the following:

        - xns_cnn: This function will use the result of the xns solver and
            create an image of the solution. The image is created via contourf.

    """

    function_name: Literal["xns_cnn"]

    #: function handler used for python functions which are loaded inside the
    #: program envelop from the outside
    _run_func: Optional[Any] = PrivateAttr(default=None)
    #: logger which is used for the internalized python functions
    _run_logger: Optional[Any] = PrivateAttr(default=None)
    #: persistent function data is not touched by releso
    _func_data: Optional[Any] = PrivateAttr(default=None)

    def __init__(self, **kwargs):
        """Internalized python function init function.

        Initializes the function and adds the observation data.

        Raises:
            ValueError: Requested function not found.
        """
        super().__init__(**kwargs)
        if self.function_name == "xns_cnn":
            from releso.util import cnn_xns_observations

            self._run_func = cnn_xns_observations.main
            self.additional_observations = (
                cnn_xns_observations.define_observation_definition()
            )
        else:  # pragma: no cover # Literal validation should catch this
            raise ValueError(
                f"The function {self.function_name} is unknown. Please check"
                " the spelling and version of the package."
            )

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """This function loads and executes the defined python file.

        The herein defined python file needs to have a function called `main`
        with three parameters.

        1. args: Namespace see SPORCommInterface
        2. logger: logger to be used in the function
        3. func_data: data variable can be used to store persistent data


        Args:
            step_information (StepReturnType):
                Previously collected step values. Should be up-to-date.
            environment_id (UUID4):
                Environment ID which is used to distinguish different
                environments which run in parallel.
            validation_id (Optional[int], optional):
                During validation the validation id signifies the validation
                episode. If none the current episode is not a validation
                episode. Defaults to None.
            core_count (int, optional):
                Can only be one since no multi processing is possible in this
                case.
            reset (bool, optional):
                Boolean on whether or not this object is called because of a
                reset. Defaults to False.

        Returns:
            StepReturnType:
                The full compliment of the step information are returned.
                The include observation, reward, done, info.
        """
        if self._first_time_setup_not_done:
            self.setup_working_directory(environment_id)
            self._first_time_setup_not_done = False

        return super().run(
            step_information, environment_id, validation_id, core_count, reset
        )


class SPORObjectExternalPythonFunction(SPORObjectPythonFunction):
    """Load a python function from an external file.

    This class is meant to load a python function from an external file and
    execute it. The function will be imported in to the package and run in the
    same python environment as the rest of the package is. If you want to call
    the function you can use the SPORCommandLineObject.

    The function needs to be called main and needs to have the
    following signature:

        args, logger, func_data

    Where args are the SPOR COMM arguments, logger is a logger provided by
    the SPORObject and, func_data is a persistent data variable which is not
    touched by the SPORObject.
    """

    python_file_path: Union[str, pathlib.Path]

    #: function handler used for python functions which are loaded inside the
    #: program envelop from the outside
    _run_func: Optional[Any] = PrivateAttr(default=None)
    #: logger which is used for the internalized python functions
    _run_logger: Optional[Any] = PrivateAttr(default=None)
    #: persistent function data is not touched by releso
    _func_data: Optional[Any] = PrivateAttr(default=None)

    @validator("python_file_path")
    @classmethod
    def validate_execution_command_path(cls, v: str):
        """Validator of the path to the python file.

        Check if path to execution command exist.

        Does not check if the file is correct or the function inside this file
        exists.

        TODO check if the correct function signature exists inside the file

        Args:
            v (str): Object to validate

        Raises:
            ParserException: If path is not correct, this error is thrown.

        Returns:
            pathlib.Path: original path if no validation error occurs.
        """
        # check if placeholder for multiprocessing is available if not use v
        # to validate directory path
        path = pathlib.Path(v)
        if path.exists() and path.is_file() and path.suffix == ".py":
            path = path.expanduser().resolve().as_posix()
        else:
            raise ParserException(
                "SPORObjectCommandline",
                "python_file_path",
                f"The path {v} does not exist.",
            )
        return path

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """This function loads and executes the defined python file.

        The herein defined python file needs to have a function called `main`
        with three parameters.

        1. args: Namespace see SPORCommInterface
        2. logger: logger to be used in the function
        3. func_data: data variable can be used to store persistent data


        Args:
            step_information (StepReturnType):
                Previously collected step values. Should be up-to-date.
            environment_id (UUID4):
                Environment ID which is used to distinguish different
                environments which run in parallel.
            validation_id (Optional[int], optional):
                During validation the validation id signifies the validation
                episode. If none the current episode is not a validation
                episode. Defaults to None.
            core_count (int, optional):
                Can only be one since no multi processing is possible in this
                case.
            reset (bool, optional):
                Boolean on whether or not this object is called because of a
                reset. Defaults to False.

        Returns:
            StepReturnType:
                The full compliment of the step information are returned.
                The include observation, reward, done, info.
        """
        if self._first_time_setup_not_done:
            self.python_file_path = pathlib.Path(self.python_file_path)
            self.setup_working_directory(environment_id)
            self.get_logger().info(
                f"Setup is performed for step with name {self.name}. For "
                f"environment with id {environment_id}."
            )

            ret_path = shutil.copy(
                self.python_file_path,
                self.save_location / self.python_file_path.name,
            )
            self.get_logger().debug(
                f"Successfully copied the python file {str(ret_path)}"
            )

            # trying to internalize the function
            try:
                sys.path.insert(0, f"{self.python_file_path.parent}{os.sep}")
                func = importlib.import_module(self.python_file_path.stem)
            except ModuleNotFoundError as err:
                raise RuntimeError(
                    f"Could not load the python file at "
                    f"{str(self.python_file_path)}"
                ) from err
            else:
                try:
                    self._run_func = func.main
                except AttributeError as err:
                    raise RuntimeError(
                        f"Could not get main function from python file"
                        f" {str(self.python_file_path)}."
                    ) from err
                else:
                    self._run_logger = set_up_logger(
                        f"spor_step_logger_{self.name.replace(' ', '_')}",
                        pathlib.Path(self.save_location / "logging"),
                        VerbosityLevel.INFO,
                        console_logging=False,
                    )
                    self.get_logger().info(
                        f"Initialized internal python function calling"
                        f" for step {self.name}."
                    )

            self._first_time_setup_not_done = False

        return super().run(
            step_information, environment_id, validation_id, core_count, reset
        )


class SPORObjectCommandLine(SPORObjectExecutor):
    """Base definition of a SPORCommandLineObject. Can be instantiated.

    Base class for all possible SPOR object classes which use the command
    line. This class is meant to add some default functionality, so that
    users do not need to add them themselves. These are:

    1. access to the (MPI) Multi-processors via
        :py:class:`ReLeSO.spor.MultiProcessor`
    2. command line command
    3. additional flags
    4. working directory
    5. whether or not to use the
        :ref:`SPOR Communication Interface <sporcominterface>`

    """

    #: Command which calls the program/script for this SPOR step.
    execution_command: str
    #: Command line options to add to the execution command. These do not
    #: include those from the communication interface these will be, if
    #: applicable, added separately.
    command_options: List[str] = Field(default_factory=list)

    @validator("execution_command")
    @classmethod
    def validate_execution_command_path(cls, v: str):
        """Validator execution_command.

        Check if path to execution command exist.

        Args:
            v (str): Object to validate

        Raises:
            ParserException: If path is not correct, this error is thrown.

        Returns:
            str: original path if no validation error occurs.
        """
        path = pathlib.Path(v)
        if path.exists():
            path = path.expanduser().resolve().as_posix()
        else:
            path = v
        if shutil.which(path) is None:
            raise ParserException(
                "SPORObjectCommandline",
                "unknown",
                f"The execution_command path {v} is not a valid executable.",
            )
        return path

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """This function runs the defined command line command.

        With the defined arguments adding if necessary multi-processing flags
        and the spor communication interface.

        Args:
            step_information (StepReturnType):
                Previously collected step values. Should be up-to-date.
            environment_id (UUID4):
                Environment ID which is used to distinguish different
                environments which run in parallel.
            validation_id (Optional[int], optional):
                During validation the validation id signifies the validation
                episode. If none the current episode is not a validation
                episode. Defaults to None.
            core_count (int, optional):
                Wanted core_count of the task to run.
                Defaults to 1.
            reset (bool, optional):
                Boolean on whether or not this object is called because of a
                reset. Defaults to False.

        Returns:
            StepReturnType:
                The full compliment of the step information are returned.
                The include observation, reward, done, info.
        """
        if self._first_time_setup_not_done:
            self.get_logger().info(
                f"Setup is performed for step with name {self.name}. For "
                f"environment with id {environment_id}."
            )
            self.setup_working_directory(environment_id)
            self._first_time_setup_not_done = False
        # first set up done

        env_logger = self.get_logger()
        step_return: Dict = {
            "observation": step_information[0],
            "reward": step_information[1],
            "done": step_information[2],
            "info": step_information[3],
        }
        # skips the step without doing anything if reset and not run_on_reset
        # add observations if the skipped step has some
        if reset and not self.run_on_reset:
            if self.additional_observations is not None:
                step_return["observation"] = self.get_default_observation(
                    step_return["observation"]
                )
        else:  # executes the step
            multi_proc_prefix = self.get_multiprocessing_prefix(
                core_count=core_count
            )
            command_list = [
                multi_proc_prefix,
                self.execution_command,
                *self.command_options,
                *self.spor_com_interface(
                    reset, environment_id, validation_id, step_information
                ),
            ]

            command = " ".join(command_list)

            # using command line to call the defined command
            exit_code, output = call_commandline(
                command, self.working_directory, env_logger
            )

            step_return["info"][self.name] = {
                "output": output,
                "exit_code": int(exit_code),
            }
            self.get_logger().debug(
                f"The return step return of the current "
                f"command is: {step_return}"
            )
            if exit_code != 0:  # error thrown during command execution
                env_logger.info("SPOR Step thrown error.")
                if self.stop_after_error:
                    env_logger.info(
                        "Will stop episode now, due to thrown error."
                    )
                    step_return["done"] = True
                    # exit code for xns if mesh is tangled hopefully
                    if (
                        self.name == "main_solver" and int(exit_code) == 137
                    ):  # pragma: no cover # this is a XNS specific case
                        step_return["info"]["reset_reason"] = (
                            f"meshTangled-{self.name}"
                        )
                    # the main_solver should always have an output only time
                    # no output is generated srun aborted early (hopefully)
                    elif (
                        self.name == "main_solver" and not output
                    ):  # pragma: no cover # this is a cluster specific case
                        step_return["info"]["reset_reason"] = (
                            f"srunError-{self.name}"
                        )
                        self.get_logger().warning(
                            "Could not find any output of failed command "
                            "assuming srun/mpi error. Trying to exit training "
                            "now."
                        )
                    else:
                        step_return["info"]["reset_reason"] = (
                            f"ExecutionFailed-{self.name}"
                        )
                step_return["reward"] = self.reward_on_error
                if self.additional_observations is not None:
                    step_return["observation"] = self.get_default_observation(
                        step_return["observation"]
                    )
            else:
                if self.use_communication_interface:
                    # if isinstance(output, dict):
                    #     self.spor_com_interface_add(output, step_return)
                    # else:
                    self.spor_com_interface_read(output, step_return)
                if self.reward_on_completion:
                    if step_return["reward"]:
                        env_logger.warning(
                            f"A reward {step_return['reward']} is already set "
                            f"but a default reward {self.reward_on_completion}"
                            " on completion is also set. The default reward "
                            "will overwrite the already set reward. Please "
                            "check if this is the wanted behavior, if not set "
                            "the reward on completion to None."
                        )
                    step_return["reward"] = self.reward_on_completion

        return (
            step_return["observation"],
            step_return["reward"],
            step_return["done"],
            step_return["info"],
        )


SPORObjectTypes = Union[
    SPORObjectCommandLine,
    SPORObjectExternalPythonFunction,
    SPORObjectInternalPythonFunction,
]


class SPORList(BaseModel):
    """The SPORList defines the custom defined steps.

    In the current version these steps take place after the FFD is
    done. These can include the fluid solver, post-processing steps,
    reward generation, additional logging, etc. In a future version the FFD
    might be incorporated into a SPOR step to stream line the internal
    functionality.

    If a step sends a stop signal all consecutive steps will be ignored and the
    Each step can generate observations, rewards and info for the current
    episode step.  These are aggregated as follows.
    """

    #: List of SPOR objects. Each SPOR object is a step  in the environment.
    steps: List[SPORObjectTypes]
    #: Definition on how the reward should be aggregated. Definition for each
    #: option is given :doc:`/SPOR`.
    reward_aggregation: Literal["sum", "min", "max", "mean", "absmax"]

    #: internal data holder for the rewards collected during a single episode.
    #: Could be local variable but is object variable to have to option to
    #: read it out later on, if need be.
    _rewards: List[RewardType] = PrivateAttr(default_factory=list)

    def get_observations(
        self,
    ) -> Optional[
        List[Union[ObservationDefinition, ObservationDefinitionMulti]]
    ]:
        """Aggregate the observations of all steps.

        Returns the aggregated observations as a flattened
        list of observations spaces coming from the SPORSteps.

        Returns:
            int: Number of observations.
        """
        observations = []
        for step in self.steps:
            fi = step.get_observations()
            if fi is None:
                continue
            if isinstance(fi, list):
                observations.extend(fi)
            else:
                observations.append(fi)
        if len(observations) == 0:
            return None
        return observations

    def _compute_reward(self) -> RewardType:
        """Aggregate and compute the reward of the SPORList.

        The aggregation method is defined in the variable
        :py:obj:`ReLeSO.spor.SPORList.reward_aggregation`.

        Raises:
            RuntimeError: If aggregation method is unknown an error is raised.

        Returns:
            RewardType: Total reward of the SPORList
        """
        if len(self._rewards) > 0:
            if self.reward_aggregation == "sum":
                return np.sum(self._rewards)
            elif self.reward_aggregation == "mean":
                return np.mean(self._rewards)
            elif self.reward_aggregation == "min":
                return np.min(self._rewards)
            elif self.reward_aggregation == "absmax":
                return self._rewards[np.argmax(np.abs(self._rewards))]
            elif self.reward_aggregation == "max":
                return np.max(self._rewards)
            else:
                err_mesg = (
                    f"The given reward aggregation method - "
                    f"{self.reward_aggregation} - is not supported."
                )
                self.get_logger().error(err_mesg)
                raise RuntimeError(err_mesg)
        else:
            self.get_logger().warning(
                "No rewards given. Setting reward to zero (0). Please check if"
                " reward should be given but did not register."
            )
            return 0

    def run(
        self,
        step_information: StepReturnType,
        environment_id: UUID4,
        validation_id: Optional[int] = None,
        core_count: int = 1,
        reset: bool = False,
    ) -> StepReturnType:
        """Run through all steps of the list and handles each output.

        When done all received observations and infos are jointly returned with
        the reward and the indicator whether or not the episode should be
        terminated.

        The values are added in-place to the step_information variable but is
        also returned.

        Args:
            step_information (StepReturnType):
                Previously collected step values. Should be up-to-date.
            environment_id (UUID4):
                Environment ID which is used to distinguish different
                environments which run in parallel.
            validation_id (Optional[int], optional):
                If currently in validation the validation needs to be
                supplied for steps which need this. If none given no validation
                is currently performed. Defaults to None.
            core_count (int, optional):
                Wanted core_count of the task to run. Defaults to 1.
            reset (bool, optional):
                Boolean on whether or not this object is called because of a
                reset. Defaults to False.

        Returns:
            StepReturnType: Results of all steps.
        """
        self._rewards = []
        observations, reward, done, info = step_information
        error_stop = False
        for step in self.steps:
            start = timer()
            if not info.get(step.name):
                info[step.name] = {}
            # error_stop should only be true if releso.SPORObject had a failure
            if error_stop:  # pragma: no cover
                if step.additional_observations:
                    # print("adding zeros")
                    observations = step.get_default_observation(observations)
                    # print(observations, observations.shape)
            else:
                try:
                    observations, reward, done, info = step.run(
                        step_information,
                        environment_id,
                        validation_id,
                        core_count,
                        reset,
                    )
                    if reward is not None:
                        self._rewards.append(reward)
                    # if info:
                    #     join_infos(info, info, self.logger_name)
                except Exception as exp:  # pragma: no cover # noqa: BLE001
                    # This should only be reached if a SPORObject has an error
                    error_message = (
                        f"The current step with name {step.name} has thrown an"
                        f" error: {exp}."
                    )
                    self.get_logger().warning(error_message, exc_info=1)
                    info[step.name]["error_reason"] = error_message
                    if step.stop_after_error:
                        done = True
                        error_stop = True
                        self.get_logger().warning(
                            f"Due to the error in the step {step.name} this"
                            " episode will now be terminated."
                        )
                        if not info.get("reset_reason"):
                            if not (
                                reset_reason := info.get(step.name).get(
                                    "reset_reason"
                                )
                            ):
                                info["reset_reason"] = error_message
                            else:
                                info[step.name]["reset_reason"] = reset_reason
            if done:
                if not info.get("reset_reason"):
                    info["reset_reason"] = "unspecified"
                    self.get_logger().warning(
                        f"Step {step.name} resulted in episode end but did not"
                        " specify reason for termination."
                    )
            reward = self._compute_reward()
            step_information = (observations, reward, done, info)
            end = timer()
            self.get_logger().debug(
                f"SPOR Step {step.name} took {end - start} seconds."
            )
        return step_information



"""Utility functions.

This files includes multiple functions and classes which are used in this
package but can not be directly associated to a single sub module.
"""
import datetime
import json
import os
import subprocess
from typing import Any, Optional

import numpy as np

from releso.util.types import InfoType


class JSONEncoder(json.JSONEncoder):
    """Encodes numpy arrays.

    This encoder class is necessary to correctly encode numpy.ndarrays, bytes
    and numpy int values.
    """

    def default(self, o: Any) -> Any:
        """Default function.

        Args:
            o (Any): To be encoded.

        Returns:
            Any: Encoded.
        """
        if isinstance(o, np.ndarray):
            return o.astype(float).tolist()
        elif isinstance(o, np.int64):
            return int(o)
        elif isinstance(o, bytes):
            return o.decode("utf-8")
        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            print(type(o), o)  # noqa: T201
            return json.JSONEncoder.default(self, "")


def which(
    program: str,
) -> Optional[str]:  # pragma: no cover # not used anymore
    """Finds if the given program is accessible or in the $PATH.

    Args:
        program (str): Program to look for

    Note:
        Verbatim taken from this stack overflow answer:
        https://stackoverflow.com/a/377028

    Returns:
        Optional[str]: None if not found else execution file.
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def call_commandline(command, folder, logger=None):
    """Executes a command which is provided as a string in the command line.

    Loosely based on a function written by Michael Binder
    (e1325632@student.tuwien.ac.at).
    """
    if logger is not None:
        logger.debug(f"Executing command {command} in {folder}")
    try:
        # try to execute the provided command in the shell
        # WARNING this is a security risk, but necessary to execute the
        # command line arguments
        output = subprocess.check_output(
            command, shell=True, cwd=folder  # noqa: S602
        )
        exitcode = 0
    except subprocess.CalledProcessError as exc:
        # if anything went wrong, catch the error, report to the user and abort
        output = exc.output
        if logger is not None:
            logger.exception(
                f"Execution failed with return code {exc.returncode}"
            )
        exitcode = exc.returncode
    return exitcode, output


def join_infos(old_info: InfoType, new_info: InfoType, logger_name: str):
    """Joins the provided old and new infos into a cohesive new dict.

    Args:
        old_info (InfoType): Already existing info dictionary which should be
        updated with the new infos.
        new_info (InfoType): Newly received info. Which needs to be added to
        the already existing infos.
        logger_name (str): LoggerName so that if necessary the log message is
        sent into the correct logging cue.
    """
    old_info.update(new_info)


def get_path_extension() -> str:
    """Creates a string which has the following characteristics.

    If not on a slurm system only the current time stamp is added
    If in a slurm job (non task array) the current time stamp + job id
    If in slurm task array job the job_id with sub-folders for each task with
    the task_id_timestamp

    If on a slurm system the slurm information are read from the environment
    variables $SLURM_JOB_ID, $SLURM_ARRAY_JOB_ID, $SLURM_ARRAY_TASK_ID.

    Returns:
        str: See function documentation body for definition.
    """
    ret_str = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    # check if slurm job is running
    if os.getenv("SLURM_CLUSTER_NAME"):  # pragma: no cover
        # check if slurm task array is running
        if (
            os.getenv("SLURM_ARRAY_TASK_COUNT")
            and (
                (tmp_var := os.getenv("SLURM_ARRAY_TASK_COUNT")) is not None
                and int(tmp_var) > 1
            )
            and os.getenv("SLURM_ARRAY_JOB_ID")
            and os.getenv("SLURM_ARRAY_TASK_ID")
        ):
            ret_str = (
                str(os.getenv("SLURM_ARRAY_JOB_ID"))
                + "/"
                + str(os.getenv("SLURM_ARRAY_TASK_ID"))
                + "_"
                + ret_str
            )
        elif os.getenv("SLURM_JOB_ID"):  # default slurm job running
            ret_str += "_" + str(os.getenv("SLURM_JOB_ID"))
    return ret_str



import logging

import pytest

from releso.util.logger import VerbosityLevel, get_parser_logger, set_up_logger


def test_get_parser_logger():
    parser_log = get_parser_logger()
    assert parser_log.name == "ReLeSO_parser"

    # check factory method
    parser_log2 = get_parser_logger()
    assert parser_log is parser_log2


def test_verbosity_level_enum():
    assert VerbosityLevel.ERROR == logging.ERROR
    assert VerbosityLevel.WARNING == logging.WARNING
    assert VerbosityLevel.INFO == logging.INFO
    assert VerbosityLevel.DEBUG == logging.DEBUG


@pytest.mark.parametrize(
    "logger_name, log_file_location, verbosity, console_logging, logger",
    [
        (None, None, None, None, None),
        (None, None, None, None, logging.getLogger("test")),
    ],
)
def test_set_up_logger(
    logger_name,
    log_file_location,
    verbosity,
    console_logging,
    logger,
    dir_save_location,
    clean_up_provider,
):
    if logger is not None:
        original_name = logger.name
    calling_dict = {
        "logger": logger,
    }
    if logger_name is not None:
        calling_dict["logger_name"] = logger_name
    if log_file_location is not None:
        calling_dict["log_file_location"] = log_file_location
    if verbosity is not None:
        calling_dict["verbosity"] = verbosity
    if console_logging is not None:
        calling_dict["console_logging"] = console_logging

    ret_logger = set_up_logger(**calling_dict)
    if logger is not None:
        assert ret_logger is logger
        assert ret_logger.name != original_name
    else:
        if logger_name is not None:
            assert ret_logger.name == logger_name

    n_handlers = 1
    if console_logging:
        n_handlers += 1
    if verbosity:
        if verbosity <= VerbosityLevel.WARNING:
            assert (dir_save_location / f"{logger_name}.log").exists()
            n_handlers += 1
    else:
        n_handlers += 1
    if console_logging:
        n_handlers += 1
    assert len(ret_logger.handlers) == n_handlers
    if log_file_location is not None:
        assert ret_logger.handlers[0].baseFilename == log_file_location
    if verbosity is not None:
        assert ret_logger.level == verbosity
    else:
        assert ret_logger.level == VerbosityLevel.INFO

    clean_up_provider(dir_save_location)



"""Internal and auxiliary plotting functions might be removed.

This file contains some plotting functions these are very limited and should
were created for the seminar and master thesis of clemens fricke. Other
plotting function exist but are not part of the library.

This file is not tested due to it being very specific to the mixd/xns use case.
"""
from typing import List

import numpy as np
from matplotlib import pyplot as plt


def get_tricontour_solution(
    width: int,
    height: int,
    dpi: int,
    coordinates: np.ndarray,
    connectivity: np.ndarray,
    solution: np.ndarray,
    sol_len: int,
    limits_min: List[float],
    limits_max: List[float],
) -> np.ndarray:  # pragma: no cover
    """_summary_.

    Args:
        width (int): _description_
        height (int): _description_
        dpi (int): _description_
        coordinates (np.ndarray): _description_
        connectivity (np.ndarray): _description_
        solution (np.ndarray): _description_
        sol_len (int): _description_
        limits_min (List[float]): _description_
        limits_max (List[float]): _description_

    Returns:
        np.ndarray: _description_
    """
    arrays: List[np.ndarray] = []
    for i in range(sol_len):
        fig = plt.figure(figsize=(width, height), dpi=dpi)
        if i == 2:  # pressure is special
            _ = plt.gca().tricontourf(
                coordinates[:, 0],
                coordinates[:, 1],
                np.clip(solution[:, i], limits_min[i], limits_max[i]) - 1,
                triangles=connectivity,
                cmap="Greys",
                vmin=limits_min[i],
                vmax=limits_max[i],
            )
        else:
            _ = plt.gca().tricontourf(
                coordinates[:, 0],
                coordinates[:, 1],
                solution[:, i],
                triangles=connectivity,
                cmap="Greys",
                vmin=limits_min[i],
                vmax=limits_max[i],
            )
        ax = plt.gca()
        ax.set_xlim((0, 1))
        ax.set_ylim((0, 1))
        ax.margins(tight=True)
        plt.axis("off")
        plt.tight_layout(pad=0, w_pad=0, h_pad=0)
        fig.canvas.draw()
        # Now we can save it to a numpy array.
        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        arrays.append(data)
        plt.close()
    arr = np.array(arrays[0])
    for i in range(sol_len - 1):
        arr[:, :, i + 1] = arrays[i + 1][:, :, 0]
    return arr


# def create_open_knot_vector(
#         degrees: List[int], n_cps: List[int]) -> List[List[float]]:
#     """Create an open knot vector according to the given parameters.

#     Args:
#         degrees (List[int]): _description_
#         n_cps (List[int]): _description_

#     Raises:
#         ValueError: _description_

#     Returns:
#         List[List[float]]: _description_
#     """
#     knot_vectors = []

#     for degree, n_cp in zip(degrees, n_cps):
#         n_middle = (degree+n_cp+1)-(2*(1+degree))
#         if n_middle < 0:
#             raise ValueError(
#                 "The n_cp needs to be high enough to actually make an open "
#                 "knot vector. Which is not the case.")
#         start = [0 for _ in range(1+degree)]
#         end = [1 for _ in range(1+degree)]
#         middle = []
#         if n_middle > 0:
#             middle = np.linspace(0, 1, n_middle+2)[1:-1].tolist()
#         start.extend(middle)
#         start.extend(end)
#         knot_vectors.append(start)
#     return knot_vectors


# def create_gif_mp4_delete_images(
#         files: List[str], save_path: str, fps: int = 5,
#         delete_used: bool = True, cut=True):
#     """_summary_.

#     Args:
#         files (List[str]): _description_
#         save_path (str): _description_
#         fps (int, optional): _description_. Defaults to 5.
#         delete_used (bool, optional): _description_. Defaults to True.
#         cut (bool, optional): _description_. Defaults to True.
#     """
#     images = []
#     for file_name in files:
#         if cut:
#             images.append(imageio.imread(file_name)[700:1300, :])
#         else:
#             images.append(imageio.imread(file_name))
#     save_path = pathlib.Path(save_path)
#     save_path.parent.mkdir(parents=True, exist_ok=True)
#     imageio.mimsave(save_path, images, fps=fps)
#     imageio.mimsave(save_path.with_suffix(".mp4"), images, fps=fps)

#     if delete_used:
#         for file_name in set(files):
#             pathlib.Path(file_name).unlink()
#         files = []


# def normalize_mesh(mesh: gus.Mesh) -> gus.Mesh:
#     """Normalize the given mesh [0,1].

#     Args:
#         mesh (gus.Mesh): Mesh to normalize.

#     Returns:
#         gus.Mesh: Normalized mesh.
#     """
#     _mesh = deepcopy(mesh)
#     min_: np.ndarray = _mesh.vertices.min(axis=0)
#     _mesh.vertices -= min_
#     max_: np.ndarray = _mesh.vertices.max(axis=0)
#     _mesh.vertices /= max_
#     return _mesh


# # TODO this is gustaf specific
# def get_face_boundaries_and_vertices(
#         mesh: gus.Mesh,
#         optional=False) -> Tuple[List[List[List[float]]], pd.DataFrame]:
#     """_summary_.

#     Args:
#         mesh (gus.Mesh): _description_
#         optional (bool, optional): _description_. Defaults to False.

#     Returns:
#         Tuple[List[List[List[float]]], pd.DataFrame]: _description_
#     """
#     # create vertices dataframe
#     vertice_df = pd.DataFrame(mesh.vertices)

#     # create plot lines for face boundaries
#     face_boundaries = []
#     for face in mesh.faces:
#         # print(face)
#         x = [vertice_df.iloc[node][0] for node in face]
#         y = [vertice_df.iloc[node][1] for node in face]
#         x.append(x[0])
#         y.append(y[0])
#         face_boundaries.append([x, y])
#     if optional:
#         bounds = []
#         for boundary in mesh.edges[mesh.outlines]:
#             x = [vertice_df.iloc[node][0] for node in boundary]
#             y = [vertice_df.iloc[node][1] for node in boundary]
#             x.append(x[0])
#             y.append(y[0])
#             bounds.append([x, y])
#         return face_boundaries, vertice_df, bounds
#     return face_boundaries, vertice_df


# def plot_mesh(
#         mesh: gus.Mesh, save_name: str, no_axis: bool = False,
#         tight: bool = False):
#     """_summary_.

#     Args:
#         mesh (gus.Mesh): _description_
#         save_name (str): _description_
#         no_axis (bool, optional): _description_. Defaults to False.
#         tight (bool, optional): _description_. Defaults to False.
#     """
#     plt.rcParams["figure.figsize"] = (20, 20)
#     plt.rcParams.update({'font.size': 22})

#     # create plotlines for faceb oundaries
#     face_boundaries, vertice_df = get_face_boundaries_and_vertices(mesh)
#     # for face in mesh.faces:
#     #   print(face)
#     #   x = [vertice_df.iloc[node][0] for node in face]
#     #   y = [vertice_df.iloc[node][1] for node in face]
#     #   x.append(x[0])
#     #   y.append(y[0])
#     #   face_boundaries.append([x,y])

#     fig, ax = plt.subplots()
#     for face in face_boundaries:
#         ax.plot(face[0], face[1], c="black")
#     ax.scatter(vertice_df[0], vertice_df[1], c="black", s=10)
#     if no_axis:
#         plt.axis('off')
#     plt.grid(True)
#     if tight:
#         plt.tight_layout()
#     plt.savefig(save_name)
#     plt.close()


# def plot_spline(
#         spline: gus.nurbs.Spline, axis: Optional[plt.Axes] = None,
#         export_path: Optional[str] = None, close: bool = False,
#         control_point_marker: str = "*",
#         control_point_marker_size: int = 120,
#         control_point_color: str = "r", spline_path_color: str = "g",
#         spline_path_alpha: float = 0.2, num_points: int = 100,
#         spline_grid_plot_step: int = 1,
#         lims: List[List[float]] = None) -> Optional[plt.Axes]:
#     """Not Implemented.

#     Args:
#         spline (gus.nurbs.Spline): _description_
#         axis (Optional[plt.Axes], optional): _description_. Defaults to None.
#         export_path (Optional[str], optional): _description_.
#           Defaults to None.
#         close (bool, optional): _description_. Defaults to False.
#         control_point_marker (str, optional): _description_. Defaults to "*".
#         control_point_marker_size (int, optional):
#         _description_. Defaults to 120.
#         control_point_color (str, optional): _description_. Defaults to "r".
#         spline_path_color (str, optional): _description_. Defaults to "g".
#         spline_path_alpha (float, optional): _description_. Defaults to 0.2.
#         num_points (int, optional): _description_. Defaults to 100.
#         spline_grid_plot_step (int, optional): _description_. Defaults to 1.
#         lims (List[List[float]], optional): _description_. Defaults to None.

#     Raises:
#         NotImplemented: _description_

#     Returns:
#         Optional[plt.Axes]: _description_
#     """
#     raise NotImplemented
#     if not axis:
#         # max 2D projection dimension
#         fig, ax = plt.subplots()
#         if lims:
#             if len(lims) > 0:
#                 ax.set_xlim(lims[0])
#                 if len(lims) > 1:
#                     ax.set_ylim(lims[1])
#                     # if len(lims) == 3:
#                     #   ax.set_zlim(lims[1])

#     mins = np.array(spline.knot_vectors).min(axis=1)
#     maxs = np.array(spline.knot_vectors).max(axis=1)
#     lines = []
#     for min_, max_ in zip(mins, maxs):
#         lines.append(np.linspace(min_, max_, num_points))
#     items_ls = []
#     for items in itertools.product(*lines):
#         items_ls.append(items)

#     items_ls = spline.evaluate(items_ls)
#     # plot grid and lines needs evaluation
#     df_items = pd.DataFrame(items_ls)
#     # currently only 2D
#     # horizontal lines
#     for i in range(0, num_points, spline_grid_plot_step):
#         ax.plot(df_items[0][i::num_points], df_items[1][i::num_points],
#                 c=spline_path_color, alpha=spline_path_alpha)
#     # horizontal lines
#     for i in range(0, num_points, spline_grid_plot_step):
#         ax.plot(
#             df_items[0][num_points*i:num_points*(i+1)],
#             df_items[1][num_points * i:num_points*(i+1)],
#             c=spline_path_color, alpha=spline_path_alpha)
#     # ax.scatter(df_items[0], df_items[1])

#     # 1D case control_mesh in spline is differently configured so this has to
#     # be implemented differently
#     if spline.para_dim_ == 1:
#         cp = spline.control_points
#         df_cp = pd.DataFrame(cp)
#         ax.scatter(df_cp[0], df_cp[1], marker=control_point_marker,
#                    c=control_point_color, s=control_point_marker_size)
#         ax.plot(df_cp[0], df_cp[1], "--", c=control_point_color)
#         evaluated_knots = spline.evaluate(
#             [[item] for item in spline.knot_vectors[0]])
#         colors = ["lightgray", "darkgray"]
#         for knot in range(len(spline.knot_vectors[0])-1):
#             ax.axvspan(evaluated_knots[knot][0], evaluated_knots[knot+1]
#                        [0], facecolor=colors[knot % 2], alpha=0.2)
#     # create polygon or cube to show boundary deformations of spline
#     elif spline.para_dim_ > 1:
#         control_mesh = spline.control_mesh_()
#         face_boundaries, vertice_df = get_face_boundaries_and_vertices(
#             control_mesh)
#         b_min = control_mesh.vertices.min(axis=0)
#         b_max = control_mesh.vertices.max(axis=0)
#     # print(b_min)
#         lines = []
#         for i in range(spline.para_dim_):
#             lines.append(np.linspace(b_min[i], b_max[i], 100))
#         control_mesh.boundaries
#         boundary_lines = []
#         for i in range(2**spline.para_dim_):
#             pass
#         min_x_yline = [[b_min[0], y] for y in y_lin]
#         min_y_xline = [[x, b_min[1]] for x in x_lin]
#         max_x_yline = [[b_max[0], y] for y in y_lin]
#         max_y_xline = [[x, b_max[1]] for x in x_lin]

#         boundary = (
#               min_x_yline+max_y_xline+max_x_yline[::-1]+min_y_xline[::-1]
#         )
#         boundary = spline.evaluate(boundary)
#         pol = Polygon(boundary, alpha=0.4)
#         ax.add_patch(pol)
#         for face in face_boundaries:
#             ax.plot(face[0], face[1], c="r")
#         ax.scatter(vertice_df[0], vertice_df[1], c="red", marker="*", s=160)
#     if export_path:
#         pathlib.Path(export_path).parent.mkdir(exist_ok=True, parents=True)
#         plt.savefig(export_path)
#     if close:
#         plt.close()



"""Non SB3 Callbacks used in ReLeSO.

Defines the custom call backs the ReLeSO library uses. Works now with
Multi-Environment training and should work with all agents.
"""

from datetime import datetime
from itertools import count
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from stable_baselines3.common.callbacks import BaseCallback


class EpisodeLogCallback(BaseCallback):
    """Episode Callback class, even for vectorized environments."""

    def __init__(
        self,
        episode_log_location: Path,
        verbose: int = 0,
        update_n_episodes: int = 1,
        logger=None,
    ):
        """Constructor for the Callback using SB3 interface.

        Args:
            episode_log_location (Path): Path to the episode log.
            verbose (int, optional): Verbosity of the callback. Defaults to 0.
            update_n_episodes (int, optional): Update the episode every n
            episodes. Defaults to 1.
            logger (optional): Logger to use for logging. If None, a default
            logger is created. Defaults to None.
        """
        super().__init__(verbose=verbose)
        self.episode_log_location: Path = episode_log_location
        self.episode_log_location.parent.mkdir(parents=True, exist_ok=True)
        self.update_n_episodes: int = update_n_episodes

        self.episodes: int = -1
        self.last_checked_episode: int = 0
        self._logger = logger
        if self._logger is None:
            from stable_baselines3.common.logger import configure

            self._logger = configure(
                str(self.episode_log_location.parent / "episode_log.log"),
                ["stdout", "csv"],
            )

        self.inter_episode_dicts: Optional[
            List[Dict[int, Union[List[float], int]]]
        ] = None

        # concatenating the dataframe should be faster than appending to it
        self.last_exported_episode: int = -1

        # pandas dataframe do not seem to actually have a method to append new
        #  data to the dataframe without knowing the index. (append exists but
        #  is deprecated) Has probably something to do with allocating new
        #  memory and pandas not wanting to actually be used for this use case.
        self.episode_rewards: List[List[float]] = []
        self.episode_n_steps: List[int] = []
        self.episode_steps_total: List[int] = []
        self.episode_wall_time: List[str] = []
        self.episode_end: List[Optional[str]] = []
        self.environment_id: List[int] = []

    def _export(self):
        """Function exports the current information into a csv file."""
        export_data_frame = pd.DataFrame({
            "steps_in_episode": self.episode_n_steps,
            "episode_reward": self.episode_rewards,
            "episode_end_reason": self.episode_end,
            "total_timesteps": self.episode_steps_total,
            "environment_id": self.environment_id,
            "wall_time": self.episode_wall_time,
        })
        # export_data_frame.to_csv(self.episode_log_location)
        # export_data_frame.reset_index(drop=True, inplace=True)
        export_data_frame.index = np.arange(
            self.last_exported_episode + 1,
            self.last_exported_episode + 1 + len(export_data_frame),
        )
        if self.last_exported_episode == -1:
            export_data_frame.to_csv(self.episode_log_location)
        else:
            export_data_frame.to_csv(
                self.episode_log_location, mode="a", header=False
            )
        self.last_exported_episode = self.episodes
        # empty lists
        self.episode_n_steps = []
        self.episode_rewards = []
        self.episode_end = []
        self.episode_steps_total = []
        self.environment_id = []
        self.episode_wall_time = []

    def _on_step(self) -> bool:
        """Function is called after a step was performed.

        Returns:
            bool: If the callback returns False, training is aborted early.
        """
        # first time setup of local vars for each environment
        if self.inter_episode_dicts is None:
            self.inter_episode_dicts: List[
                Dict[str, Union[List[float], int]]
            ] = []
            for _ in range(len(self.locals["dones"])):
                self.inter_episode_dicts.append({
                    "current_episode_rewards": [],
                    "last_start_step": 0,
                })
        continue_training = True
        for idx, loc_vars, done, reward, info in zip(
            count(),
            self.inter_episode_dicts,
            self.locals["dones"],
            self.locals["rewards"],
            self.locals["infos"],
        ):
            loc_vars["current_episode_rewards"].append(reward)
            if done:
                self.episodes += 1
                self.episode_rewards.append(
                    sum(loc_vars["current_episode_rewards"])
                )
                loc_vars["current_episode_rewards"] = []
                self.episode_n_steps.append(
                    self.n_calls - loc_vars["last_start_step"]
                )
                loc_vars["last_start_step"] = self.n_calls + 1
                self.episode_steps_total.append(self.num_timesteps)
                self.episode_wall_time.append(str(datetime.now()))
                reset_reason = (
                    None
                    if "reset_reason" not in info
                    else info["reset_reason"]
                )
                self.episode_end.append(reset_reason)
                self.environment_id.append(idx)
                if reset_reason == "srunError-main_solver":  # pragma: no cover
                    continue_training = False
        if any(
            episode % self.update_n_episodes == 0
            for episode in range(self.last_checked_episode, self.episodes)
        ):
            self._export()
        self.last_checked_episode = self.episodes
        return continue_training

    def _on_training_end(self) -> None:
        """Function is called when training is terminated."""
        self._export()


class StepLogCallback(BaseCallback):
    """Step Callback class.

    This class tracks all step-wise information that might come in handy
    during evaluation. Originally created to easily track the actions
    undertaken in each episode for better evaluation of the learned policy.
    """

    def __init__(
        self,
        step_log_location: Path,
        verbose: int = 0,
        update_n_steps: int = 0,
        log_infos: bool = False,
        logger=None,
    ):
        """Constructor for the Callback using SB3 interface.

        Args:
            step_log_location (Path): Path to the step log file.
            verbose (int, optional): Verbosity of the callback. Defaults to 0.
            update_every (int, optional): Update the step log file every n
            steps. Defaults to 0 which triggers the update after every episode.
            log_infos (bool, optional): Whether to log the infos. Defaults to
            False.
            logger (optional): Logger to use for logging. If None, a default
            logger is created. Defaults to None.
        """
        super().__init__(verbose)
        self.step_log_location: Path = step_log_location
        self.step_log_location.parent.mkdir(parents=True, exist_ok=True)
        self.current_episodes: list[int] = []
        self.update_n_episodes: int = update_n_steps
        self.first_export: bool = True
        self.current_max_episodes: int = 0
        self.log_infos: bool = log_infos

        self._logger = logger
        if self._logger is None:
            from stable_baselines3.common.logger import configure

            self._logger = configure(
                str(self.step_log_location.parent / "episode_log.log"),
                ["stdout", "csv"],
            )
        self._logger.info(
            (
                "The StepLogCallback should only be used for debugging purposes"
                " and only for use-cases with a smallish observation space."
                " Otherwise it will create a lot of data and slow down the"
                " training process. Use with caution!"
            )
        )
        self._reset_internal_storage()

    def _reset_internal_storage(self) -> None:
        """Reset the internally used lists which store the step-wise
        information since last updating the logfile.
        """
        self.timesteps = []  # Store step numbers
        self.episodes = []  # Store episode numbers
        self.dones = []  # Store if the episode has ended
        self.prev_obs = []  # Store observations at the start of time step
        self.actions = []  # Store actions
        self.new_obs = []  # Store observations after completion of time step
        self.rewards = []  # Optionally store rewards
        self.infos = []  # Optionally store additional information

    def _export(self) -> None:
        """Convert the step-wise information to a dataframe and export to csv."""
        # Combine all relevant information into a pandas DataFrame
        export_data_frame = pd.DataFrame({
            "episodes": self.episodes,
            "dones": self.dones,
            "prev_obs": self.prev_obs,
            "actions": self.actions,
            "new_obs": self.new_obs,
            "rewards": self.rewards,
        })
        if self.log_infos:
            export_data_frame["infos"] = self.infos
        export_data_frame.index = self.timesteps
        export_data_frame.index.name = "timesteps"
        # Write the data to file
        export_data_frame.to_json(
            self.step_log_location,
            orient="records",
            lines=True,
            mode="a" if not self.first_export else "w",
        )
        # data frame has been exported already at least once, so reset the flag
        self.first_export = False
        # reset the internal storage
        self._reset_internal_storage()

    def _on_step(self) -> bool:
        """Function that is called after a step was performed.

        Returns:
            bool: If the callback returns False, training is aborted early.
        """
        # Retrieve the step-wise information that we want to keep track of
        dones = self.locals[
            "dones"
        ]  # Field indicating if the episode has ended
        if len(self.current_episodes) == 0:
            n_dones = len(dones)
            self.current_episodes = [idx - n_dones for idx in range(n_dones)]
            self.current_max_episodes = max(self.current_episodes)
        prev_obs = self.model._last_obs  # Old observations
        actions = self.locals["actions"]  # Agent's actions
        new_obs = self.locals["new_obs"]  # New observations
        rewards = self.locals["rewards"]  # Rewards (optional)
        infos = self.locals["infos"]  # Additional information

        # Store additional information
        self.timesteps.append(self.num_timesteps)
        self.episodes.append(self.current_episodes)
        self.dones.append(dones)
        self.prev_obs.append(prev_obs)
        self.actions.append(actions)
        self.new_obs.append(new_obs)
        self.rewards.append(rewards)

        if self.log_infos:
            infos = self.locals["infos"]  # Additional information
            self.infos.append(infos)

        done_export = False
        # Check if the environment has completed an episode
        for idx, done in enumerate(dones):
            if done:
                # If the environment has completed an episode, locals["new_obs"]
                # will already contain the observation AFTER a reset has been
                # performed and not the original terminal observation with which the
                # episode ended, so overwrite the entries with the true terminal
                # observation
                self.new_obs[-1][idx] = self.locals["infos"][idx][
                    "terminal_observation"
                ]
                # If the update is supposed to be performed after an episode has
                # been completed ...
                if self.update_n_episodes == 0:
                    # schedule the export of the information
                    done_export = True
                # Always increase the episode counter
                self.current_episodes[idx] = self.current_max_episodes + 1
                self.current_max_episodes = self.current_episodes[idx]
        # If export scheduled or export with the given
        # frequency
        if done_export or (
            self.update_n_episodes != 0
            and any(
                timestep % self.update_n_episodes == 0
                for timestep in self.timesteps
            )
        ):
            self._export()

        return True



"""File defines the exceptions of the ReLeSO toolbox/library."""
import enum
import logging
import os
import platform
import sys
from typing import List

from releso.util.logger import get_parser_logger


class Stylings(enum.Enum):
    """Possible stylings supported by this Tool."""

    Bold = 1
    Faint = 2
    Italic = 3
    Underline = 4
    Black = 30
    Red = 31
    Green = 32
    Yellow = 33
    Blue = 34
    Magenta = 35
    Cyan = 36
    White = 37

    def __str__(self) -> str:
        """Encode enum value into a string.

        Returns:
            str: encoded enum value.
        """
        return "%s" % self.value


def check_if_color_is_supported_in_console() -> bool:
    """Checks if the connected consol has color support.

    Original from this gist https://gist.github.com/ssbarnea/1316877.

    Returns:
        bool: Terminal supports color.
    """
    for handle in [sys.stdout, sys.stderr]:
        if (hasattr(handle, "isatty") and handle.isatty()) or (
            "TERM" in os.environ and os.environ["TERM"] == "ANSI"
        ):
            if platform.system() == "Windows" and not (
                "TERM" in os.environ and os.environ["TERM"] == "ANSI"
            ):
                return False
            else:
                return True
        else:
            return False
    return False


#: boolean value on whether or not coloring is supported
color_supported = check_if_color_is_supported_in_console()


def output_styling(message: str, stylings: List[Stylings]) -> str:
    """Style the string according to the given styles.

    If ANSI styling is available the message is styled by an the defined
    styling. Only the first 3 Stylings are actually used the others will be
    ignored.

    Valid stylings can be found in the enum: ReLeSO.exceptions.Stylings

    Args:
        message (str): string to style
        stylings (List[Stylings]): Stylings to apply to the message.

    Returns:
        (str): ANSI encoded styled message.
    """
    return (
        f"\033[{';'.join(map(str, stylings[:3]))}m{message}\033[0m"
        if color_supported
        else message
    )


def red(message: str) -> str:
    """Color string red.

    If ANSI styling is available the message is colored red and bold. After the
    message the styling is removed again.

    For more information please see ReLeSO.exceptions.output_styling().

    Args:
        message (str): string to color

    Returns:
        (str): ANSI encoded colored message.
    """
    return output_styling(
        message=message, stylings=[Stylings.Red, Stylings.Bold]
    )


def underline(message: str) -> str:
    """Underline the string.

    If ANSI styling is available the message is styled by an underscore and
    bold. After the message the styling is removed again.

    For more information please see ReLeSO.exceptions.output_styling().

    Args:
        message (str): string to strike through

    Returns:
        (str): ANSI encoded struck through message.
    """
    return output_styling(
        message=message, stylings=[Stylings.Underline, Stylings.Bold]
    )


class ParserException(ValueError):
    """Parser Exception for the ReLeSO package.

    Shows the context of the error. Also colors the output if coloring is
    available.

    Uses the ValueError base class for compatibility with the pydantic
    validation engine.

    Args:
        parent (str): Parent items names where the error occurred.
        item (str): Name of the item which caused the error.
        message (str): Custom message to display in the Exception.
    """

    def __init__(self, parent: str, item: str, message: str) -> None:
        """Parser Exception constructor.

        Args:
            parent (str): Parent items names where the error occurred.
            item (str): Name of the item which caused the error.
            message (str): Custom message to display in the Exception.
        """
        get_parser_logger().exception(
            f"In {parent} object while parsing {item} "
            f"the following error has occurred: {message}"
        )
        super().__init__(
            f"In {underline(parent)} object while parsing {underline(item)} "
            f"the following error has occurred: {red(message)}"
        )


class AgentUnknownException(Exception):
    """Agent unknown exception for the ReLeSO package.

    Thrown when an agent is unknown. Configurable logger.

    Args:
        agent (str): Name of the agent which is not set.
        logger (str, optional): Name of the logger.
        Defaults to "ReLeSO_environment".
    """

    def __init__(self, agent: str, logger: str = "ReLeSO_environment") -> None:
        """Constructor AgentUnknownException.

        Args:
            agent (str): Name of the agent which is not set.
            logger (str, optional): Name of the logger.
            Defaults to "ReLeSO_environment".
        """
        mes_str: str = f"The {agent} is unknown. Check the spelling or"
        " put in a request to add this agent."
        logging.getLogger(logger).exception(mes_str)
        super().__init__(mes_str)


class ValidationNotSet(Exception):
    """Validation not set exception.

    Thrown when a validation
    environment is needed but can not be created due to un-configured
    validation settings. Configurable logger.

    Args:
        logger (str):
            Name of the logger where the exception should be logged over.
            Defaults to: 'ReLeSO_environment'
    """

    def __init__(self, logger: str = "ReLeSO_validation_environment") -> None:
        """Constructor ValidationNotSetException.

        Args:
            logger (str, optional): Name of the logger.
            Defaults to "ReLeSO_validation_environment".
        """
        mes_str: str = (
            "Could not create a validation environment due to"
            "unavailability of validation parameters. Please add the "
            "validation parameters to the json file."
        )
        logging.getLogger(logger).exception(
            type(self).__name__ + ": " + mes_str
        )
        super().__init__(mes_str)



"""File defines the FeatureExtractors currently implemented in this framework.

These are implemented in a more or less slapdash fashion. Use at own Risk.
"""
import logging
from typing import Literal, Optional

import torch as th
from gymnasium import Space, spaces
from stable_baselines3.common.preprocessing import (
    get_flattened_obs_dim,
    is_image_space,
)
from stable_baselines3.common.torch_layers import BaseFeaturesExtractor
from stable_baselines3.common.type_aliases import TensorDict
from torchvision import models, transforms


class FeatureExtractor(BaseFeaturesExtractor):
    """Normal Feature extractor class implementation for ReLeSO.

    The standard feature extractor can only be used with a Box defined
    Observation space. For more complex Observations spaces see
    :py:class:`ReLeSO.feature_extractor.CombinedExtractor`
    """

    def __init__(
        self,
        observation_space: Space,
        features_dim: int = 128,
        without_linear: bool = False,
        network_type: Literal["resnet18", "mobilenet_v2"] = "resnet18",
        logger: Optional[logging.Logger] = None,
    ):
        """Constructor for normal feature extractor.

        The feature extractor is used to also be able to handle image based
        observations better and to give the option to used pretrained networks
        as the feature extractor. Multiple are available please look into the
        code to check the currently available pretrained networks. Not all are
        listed in the network_type variable hint.

        Args:
          observation_space (Space): Observations space of the environment.
          cnn_output_dim (int, optional): How many features the feature
            extractor should return. Defaults to 128.
          features_dim (int, optional): Number of dimensions in the feature
            dimension.
          without_linear (bool, optional): Use the pretrained nets without a
            linear layer at the end. Setting the output_dim is not possible
            since the output sie is directly dependent on the input size.
            Defaults to False.
          network_type (Literal["resnet18", "mobilenetv2"], optional):Network
            to use for the cnn extractor. Defaults to "resnet18".
          logger (Optional[logging.Logger], optional): Logger to use for
            logging purposes. Defaults to None.

        """
        super().__init__(observation_space, features_dim=features_dim)
        self.without_linear = without_linear
        # Define the network
        if network_type == "resnet18":
            pre_network = models.resnet18(
                weights=models.ResNet18_Weights.IMAGENET1K_V1
            )
            self.model = th.nn.Sequential(
                *[
                    transforms.Normalize(
                        (0.485, 0.456, 0.406), (0.229, 0.224, 0.225)
                    ),
                    pre_network.conv1,
                    pre_network.bn1,
                    pre_network.relu,
                    pre_network.maxpool,
                    pre_network.layer1,
                    pre_network.layer2,
                    pre_network.layer3,
                    pre_network.layer4,
                    th.nn.Flatten(),
                ]
            )
        elif network_type == "mobilenet_v2":
            pre_network = models.mobilenet_v2(
                weights=models.MobileNet_V2_Weights.IMAGENET1K_V1
            )
            self.model = th.nn.Sequential(
                *[
                    transforms.Normalize(
                        (0.485, 0.456, 0.406), (0.229, 0.224, 0.225)
                    ),
                    pre_network.features,
                    th.nn.Flatten(),
                ]
            )
        # elif network_type == "mobilenetv3_small":
        #     pre_network = models.mobilenet_v3_small(pretrained=True)
        #     self.model = th.nn.Sequential(
        #         *[
        #             transforms.Normalize(
        #                 (0.485, 0.456, 0.406), (0.229, 0.224, 0.225)
        #             ),
        #             pre_network.features,
        #             th.nn.Flatten(),
        #         ]
        #     )
        # elif network_type == "mobilenetv3_large":
        #     pre_network = models.mobilenet_v3_large(pretrained=True)
        #     self.model = th.nn.Sequential(
        #         *[
        #             transforms.Normalize(
        #                 (0.485, 0.456, 0.406), (0.229, 0.224, 0.225)
        #             ),
        #             pre_network.features,
        #             th.nn.Flatten(),
        #         ]
        #     )
        # elif network_type == "inception_v3":
        #     # extracting the linear layer is not possible. But can be
        #     # overwritten see below.
        #     if self.without_linear:
        #         if logger:
        #             logger.error(
        #                 "FeatureExtractor: Inception_v3 can only be used "
        #                 "with linear layer."
        #             )
        #         raise RuntimeError(
        #             "FeatureExtractor: Inception_v3 can only be used with "
        #             "linear layer."
        #         )
        #     self.model = models.inception_v3(pretrained=True)
        else:
            if logger:
                logger.error(
                    f"The given network type of {network_type} is unknown. "
                    "Please choose one that is known."
                )
            raise RuntimeError(
                f"FeatureExtractor: Given network type -{network_type}- "
                "unknown."
            )
        # deactivate training
        for param in self.model.parameters():
            param.requires_grad = False

        # Compute shape by doing one forward pass
        with th.no_grad():
            n_flatten = self.model(
                th.as_tensor(observation_space.sample()[None]).float()
            ).shape
        if self.without_linear:
            features_dim = n_flatten[1]
        else:
            # if network_type == "inception_v3":
            #     # inception_v3 has the linear layer inside the model in a way
            #     # that it is not easy to separate it from the convolutional
            #     # layers. Therefor need to overwrite the model's linear layer
            #     self.model.fc = th.nn.Linear(
            #         self.model.fc.in_features, features_dim, th.nn.ReLU()
            #     )
            #     # only so that forward correctly works. Actually uses the
            #     # linear layer but it is included inside the model itself.
            #     self.without_linear = True
            # else:
            self.linear = th.nn.Sequential(
                th.nn.Linear(n_flatten[1], features_dim), th.nn.ReLU()
            )
        if logger:
            logger.warning(
                f"FeatureExtractor: Used pretrained {network_type} results in "
                f"{features_dim} feature dims."
            )
        self._features_dim = features_dim

    def forward(self, observations: th.Tensor) -> th.Tensor:
        """Forward pass of the Network defined.

        Args:
            observations (th.Tensor): Input for the network

        Returns:
            th.Tensor: Output of the network
        """
        observations = self.model(observations)
        # return observations
        if not self.without_linear:
            observations = self.linear(observations)
        return observations


class CombinedExtractor(BaseFeaturesExtractor):
    """Combined Extractor can use a Dict definition of the observations space.

    Notes: Class is a direct copy from
        stable_baselines3.common.torch_layers.CombinedExtractor. Only change is
        the image feature extractor.
    """

    def __init__(
        self,
        observation_space: spaces.Dict,
        cnn_output_dim: int = 256,
        without_linear: bool = False,
        network_type: Literal["resnet18", "mobilenetv2"] = "resnet18",
        logger: Optional[logging.Logger] = None,
    ):
        """Combined Feature extractor constructor.

        The feature extractor is used to also be able to handle image based
        observations better and to give the option to used pretrained
        networks as the feature extractor. Multiple are available please look
        into the code to check the currently available pretrained networks.
        Not all are listed in the network_type variable hint.

        The combined feature extractor extends this capability to non uniform
        observations space definitions. For example if image based
        observations are mixed with standard (scalar) observations. Or
        multiple images are used as an observation.

        Args:
            observation_space (spaces.Dict): Observations space of the
            environment.
            cnn_output_dim (int, optional): How many features the feature
            extractor should return. Defaults to 256.
            without_linear (bool, optional): Use the pretrained nets without a
            linear layer at the end. Setting the output_dim is not possible
            since the output is directly dependent on the input size.
            Defaults to False.
            network_type (Literal["resnet18", "mobilenetv2"],optional): Network
            to use for the cnn extractor. Defaults to "resnet18".
            logger (Optional[logging.Logger], optional): Logger to use for
            logging purposes. Defaults to None.
        """
        # TODO we do not know features-dim here before going over all the
        # items, so put something there. This is not clean!
        super().__init__(observation_space, features_dim=1)

        extractors = {}

        total_concat_size = 0
        for key, subspace in observation_space.spaces.items():
            if is_image_space(subspace):
                extractors[key] = FeatureExtractor(
                    observation_space=subspace,
                    features_dim=cnn_output_dim,
                    without_linear=without_linear,
                    network_type=network_type,
                    logger=logger,
                )
                total_concat_size += extractors[key]._features_dim
            else:
                # The observation key is a vector, flatten it if needed
                extractors[key] = th.nn.Flatten()
                total_concat_size += get_flattened_obs_dim(subspace)

        self.extractors = th.nn.ModuleDict(extractors)

        # Update the features dim manually
        self._features_dim = total_concat_size

    def forward(self, observations: TensorDict) -> th.Tensor:
        """Forward pass of the Network defined.

        Args:
            observations (th.Tensor): Input for the network

        Returns:
            th.Tensor: Output of the network
        """
        encoded_tensor_list = []

        for key, extractor in self.extractors.items():
            encoded_tensor_list.append(extractor(observations[key]))
        return th.cat(encoded_tensor_list, dim=1)



import pytest
from pydantic import ValidationError
from stable_baselines3 import PPO

from releso.validation import Validation
from releso.verbosity import VerbosityLevel


@pytest.mark.parametrize(
    [
        "validation_freq",
        "validation_values",
        "save_best_agent",
        "validate_on_training_end",
        "max_timesteps_in_episode",
        "end_episode_on_geometry_not_changed",
        "reward_on_geometry_not_changed",
        "reward_on_episode_exceeds_max_timesteps",
        "val_error",
        "val_warning",
    ],
    [
        (
            1,
            [],
            False,
            False,
            None,
            None,
            None,
            None,
            "at least 1 item",
            False,
        ),
        (1, [12], False, False, None, None, None, None, False, False),
        (
            1,
            [12],
            False,
            False,
            None,
            True,
            None,
            None,
            False,
            "Please set a reward value for geometry not changed",
        ),
        (
            1,
            [12],
            False,
            False,
            5,
            None,
            None,
            None,
            False,
            "Please set a reward value for max time steps exceeded",
        ),
        (
            1,
            [12],
            False,
            False,
            None,
            None,
            -1,
            None,
            "Reward can only be set if end_episode_on_geometry_not_changed",
            False,
        ),
        (
            1,
            [12],
            False,
            False,
            None,
            None,
            None,
            -1,
            "Reward can only be set if max_timesteps_in_episode is a",
            False,
        ),
    ],
)
def test(
    validation_freq,
    validation_values,
    save_best_agent,
    validate_on_training_end,
    max_timesteps_in_episode,
    end_episode_on_geometry_not_changed,
    reward_on_geometry_not_changed,
    reward_on_episode_exceeds_max_timesteps,
    val_error,
    val_warning,
    dir_save_location,
    caplog,
):
    calling_dict = {
        "validation_freq": validation_freq,
        "validation_values": validation_values,
        "save_best_agent": save_best_agent,
        "validate_on_training_end": validate_on_training_end,
        "save_location": dir_save_location,
    }
    if max_timesteps_in_episode is not None:
        calling_dict["max_timesteps_in_episode"] = max_timesteps_in_episode
    if end_episode_on_geometry_not_changed is not None:
        calling_dict[
            "end_episode_on_geometry_not_changed"
        ] = end_episode_on_geometry_not_changed
    if reward_on_geometry_not_changed is not None:
        calling_dict[
            "reward_on_geometry_not_changed"
        ] = reward_on_geometry_not_changed
    if reward_on_episode_exceeds_max_timesteps is not None:
        calling_dict[
            "reward_on_episode_exceeds_max_timesteps"
        ] = reward_on_episode_exceeds_max_timesteps

    if val_error:
        with pytest.raises(ValidationError) as err:
            Validation(**calling_dict)
        assert val_error in str(err.value)
        return
    with caplog.at_level(VerbosityLevel.WARNING):
        validation = Validation(**calling_dict)
    if val_warning:
        assert val_warning in caplog.text
    if end_episode_on_geometry_not_changed is None:
        end_episode_on_geometry_not_changed = False
    if (
        max_timesteps_in_episode is not None
        and reward_on_episode_exceeds_max_timesteps is None
    ):
        reward_on_episode_exceeds_max_timesteps = 0
    if (
        end_episode_on_geometry_not_changed
        and reward_on_geometry_not_changed is None
    ):
        reward_on_geometry_not_changed = 0

    assert validation.should_add_callback()
    validation.validation_freq = -1
    assert not validation.should_add_callback()

    val_environment_parameters = (
        validation.get_environment_validation_parameters()
    )
    assert (
        val_environment_parameters["max_timesteps_in_episode"]
        == max_timesteps_in_episode
    )
    assert (
        val_environment_parameters["end_episode_on_geometry_not_changed"]
        == end_episode_on_geometry_not_changed
    )
    assert (
        val_environment_parameters["reward_on_geometry_not_changed"]
        == reward_on_geometry_not_changed
    )
    assert (
        val_environment_parameters["reward_on_episode_exceeds_max_timesteps"]
        == reward_on_episode_exceeds_max_timesteps
    )
    assert val_environment_parameters["validation_values"] == validation_values


@pytest.mark.parametrize("save_location", [(False), (True)])
def test_validation_with_environment(
    save_location, dir_save_location, provide_dummy_environment
):
    calling_dict = {
        "validation_freq": 1,
        "validation_values": [12],
        "save_best_agent": save_location,
        "validate_on_training_end": False,
        "save_location": dir_save_location,
    }
    validation = Validation(**calling_dict)
    assert validation.should_add_callback()
    variable_dict = {
        "eval_environment": provide_dummy_environment,
    }
    if save_location:
        variable_dict["save_location"] = dir_save_location
    call_back = validation.get_callback(**variable_dict)
    assert call_back.eval_freq == 1
    if save_location:
        assert str(dir_save_location) in str(call_back.best_model_save_path)
    else:
        assert call_back.best_model_save_path is None

    agent = PPO("MlpPolicy", provide_dummy_environment, verbose=0, n_steps=100)

    validation.end_validation(agent, provide_dummy_environment)



import datetime

import gymnasium as gym
import pytest
from pydantic import BaseModel
from stable_baselines3 import A2C, DDPG, DQN, PPO, SAC

from releso.agent import (
    A2CAgent,
    AgentTypeDefinition,
    BaseAgent,
    BaseTrainingAgent,
    DDPGAgent,
    DQNAgent,
    PPOAgent,
    PretrainedAgent,
    SACAgent,
)
from releso.feature_extractor import CombinedExtractor, FeatureExtractor
from releso.verbosity import VerbosityLevel


@pytest.mark.parametrize(
    "log_name",
    [
        (None),
        (False),
        ("test"),
    ],
)
def test_base_agent_tensorboard_log(log_name, dir_save_location):
    calling_dict = {
        "save_location": dir_save_location,
    }
    if log_name is not False:
        calling_dict["tensorboard_log"] = log_name

    b_agent = BaseAgent(**calling_dict)

    if log_name:
        assert b_agent.get_next_tensorboard_experiment_name() == (
            f"{log_name}_"
            f"{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
        )
    else:
        assert b_agent.get_next_tensorboard_experiment_name() is None


@pytest.mark.parametrize(
    "policy, use_custom_feature_extractor, cfe_without_linear, policy_kwargs",
    [
        ("MlpPolicy", None, False, None),
        ("MlpPolicy", None, True, {}),
        ("CnnPolicy", None, False, None),
        ("CnnPolicy", None, True, {}),
        ("MultiInputPolicy", None, False, {}),
        ("MultiInputPolicy", None, True, {}),
        ("MlpPolicy", "resnet18", True, {}),
        ("MlpPolicy", "resnet18", False, {}),
        ("CnnPolicy", "resnet18", False, {}),
        ("CnnPolicy", "resnet18", True, {}),
        ("MultiInputPolicy", "resnet18", False, {}),
        ("MultiInputPolicy", "resnet18", True, {}),
        ("MlpPolicy", "mobilenet_v2", True, {}),
        ("MlpPolicy", "mobilenet_v2", False, {}),
        ("CnnPolicy", "mobilenet_v2", False, {}),
        ("CnnPolicy", "mobilenet_v2", True, {}),
        ("MultiInputPolicy", "mobilenet_v2", False, {}),
        ("MultiInputPolicy", "mobilenet_v2", True, {}),
    ],
)
def test_base_training_agent(
    policy,
    use_custom_feature_extractor,
    cfe_without_linear,
    policy_kwargs,
    dir_save_location,
    caplog,
):
    calling_dict = {
        "save_location": dir_save_location,
        "policy": policy,
        "cfe_without_linear": cfe_without_linear,
        "policy_kwargs": policy_kwargs,
    }
    if use_custom_feature_extractor is not None:
        calling_dict[
            "use_custom_feature_extractor"
        ] = use_custom_feature_extractor
    base_training_agent = BaseTrainingAgent(**calling_dict)

    with caplog.at_level(VerbosityLevel.WARNING):
        additional_kwargs = base_training_agent.get_additional_kwargs()
    assert isinstance(additional_kwargs, dict)
    assert additional_kwargs["policy"] == policy
    if policy == "MlpPolicy" and use_custom_feature_extractor is not None:
        assert (
            "Please use the CnnPolicy or the MultiInputPolicy with"
            in caplog.text
        )
        assert (
            additional_kwargs["policy_kwargs"]["features_extractor_class"]
            == FeatureExtractor
        )
    if policy == "CnnPolicy" and use_custom_feature_extractor is not None:
        assert (
            additional_kwargs["policy_kwargs"]["features_extractor_class"]
            == FeatureExtractor
        )
    if (
        policy == "MultiInputPolicy"
        and use_custom_feature_extractor is not None
    ):
        assert (
            additional_kwargs["policy_kwargs"]["features_extractor_class"]
            == CombinedExtractor
        )
    if use_custom_feature_extractor is not None:
        if cfe_without_linear:
            assert (
                additional_kwargs["policy_kwargs"][
                    "features_extractor_kwargs"
                ]["without_linear"]
                is True
            )
        assert (
            additional_kwargs["policy_kwargs"]["features_extractor_kwargs"][
                "network_type"
            ]
            == use_custom_feature_extractor
        )
        assert (
            "logger"
            in additional_kwargs["policy_kwargs"][
                "features_extractor_kwargs"
            ].keys()
        )


class Dummy(BaseModel):
    agent: AgentTypeDefinition


@pytest.mark.parametrize("normalizer_divisor", [(1), (100), (0)])
@pytest.mark.parametrize(
    "agent, wanted_agent, resulting_agent",
    [
        ("PPO", PPOAgent, PPO),
        ("A2C", A2CAgent, A2C),
        ("DDPG", DDPGAgent, DDPG),
        ("DQN", DQNAgent, DQN),
        ("SAC", SACAgent, SAC),
    ],
)
def test_agents(
    agent,
    normalizer_divisor,
    wanted_agent,
    resulting_agent,
    dir_save_location,
    caplog,
):
    calling_dict = {
        "save_location": dir_save_location,
        "type": agent,
        "policy": "MlpPolicy",
    }
    ag = Dummy(agent=calling_dict)
    assert isinstance(ag.agent, wanted_agent)

    with caplog.at_level(VerbosityLevel.WARNING):
        if agent == "DQN":  # must use discrete actions
            agent = ag.agent.get_agent(
                gym.make("MountainCar-v0"), normalizer_divisor
            )
        else:
            agent = ag.agent.get_agent(
                gym.make("Pendulum-v1"), normalizer_divisor
            )
        if agent in ["PPO", "A2C"] and normalizer_divisor == 0:
            assert "Normalizer divisor is 0, will use 1." in caplog.text
            normalizer_divisor = 1

    assert isinstance(agent, resulting_agent)


@pytest.mark.parametrize(
    "agent, tensorboard_run_directory, tensorboard_log",
    [
        ("PPO", None, None),
        ("PPO", "test", None),
        ("PPO", None, "test"),
    ],
)
def test_pretrained_agent(
    agent,
    tensorboard_run_directory,
    tensorboard_log,
    dir_save_location,
    dummy_file,
):
    calling_dict = {
        "save_location": dir_save_location,
        "type": agent,
        "tensorboard_log": tensorboard_log,
        "path": dummy_file,
    }
    if tensorboard_run_directory:
        calling_dict["tesorboard_run_directory"] = tensorboard_run_directory
    ag = PretrainedAgent(**calling_dict)
    if tensorboard_run_directory:
        assert (
            ag.get_next_tensorboard_experiment_name()
            == tensorboard_run_directory
        )
    else:
        if tensorboard_log is not None:
            assert (
                ag.get_next_tensorboard_experiment_name()
                == f"{tensorboard_log}_"
                f"{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
            )
        else:
            assert ag.get_next_tensorboard_experiment_name() is None


@pytest.mark.skip(reason="no way of easily testing this")
def test_pretrained_load_correct_agent():
    pass



"""File holds the definition class for the validation."""
import pathlib
from typing import Any, Dict, Optional, Tuple

from pydantic.class_validators import validator
from pydantic.types import conint, conlist
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.callbacks import EvalCallback
from stable_baselines3.common.evaluation import evaluate_policy
from stable_baselines3.common.type_aliases import GymEnv

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.util.logger import get_parser_logger


class Validation(BaseModel):
    """Parser class to define the validation to be performed during training.

    This class is used for the configuration of how validation is to be
    performed during training.
    """

    #: How many timesteps should pass by learning between validation runs
    validation_freq: Optional[conint(ge=1)]
    #: List of validation items. This will be revised later on #TODO
    validation_values: conlist(float, min_items=1)
    #: Whether or not to save the best agent. If agent is not saved only
    #: results will be reported by the agent which produced them will not be
    #: saved.
    save_best_agent: bool
    #: Whether or not to also run a validation when the training is terminated.
    validate_on_training_end: bool
    #: after how many timesteps inside a single episode should the episode be
    #: terminated.
    max_timesteps_in_episode: Optional[conint(ge=1)] = None
    #: Should the episode be terminated if the geometry representation has not
    #: changed between timesteps?
    end_episode_on_geometry_not_changed: bool = False
    #: What reward should be added to the step reward if the geometry was not
    #: changed for the defined number of steps.
    reward_on_geometry_not_changed: Optional[float] = None
    #: What reward should be added to the step reward if the maximal timesteps
    #: per episode is exceeded.
    reward_on_episode_exceeds_max_timesteps: Optional[float] = None

    @validator("reward_on_geometry_not_changed", always=True)
    @classmethod
    def check_if_reward_given_if_geometry_not_change_episode_killer_activated(
        cls, value: float, values: Dict[str, Any]
    ) -> float:
        """Validator for reward_on_geometry_not_changed.

        Checks that 1) if a reward is set, also the boolean value for the
        end_episode_on_geometry_not_changed is True.
        2) If end_episode_on_geometry_not_changed is True a reward value is set

        Args:
            value (float): value to validate
            values (Dict[str, Any]): previously validated values
            (here end_episode_on_geometry_not_changed is important)

        Raises:
            ParserException: Error is thrown if one of the conditions is
            not met.

        Returns:
            float: reward for the specified occurrence.
        """
        # due to pydantic this first statement should never be able to happen
        if (
            "end_episode_on_geometry_not_changed" not in values
        ):  # pragma: no cover
            raise ParserException(
                "Validation",
                "reward_on_geometry_not_changed",
                "Could not find definition of parameter "
                "end_episode_on_geometry_not_changed, please define this "
                "variable since otherwise this variable would have no "
                "function.",
            )
        if value is not None and (
            values["end_episode_on_geometry_not_changed"] is None
            or not values["end_episode_on_geometry_not_changed"]
        ):
            raise ParserException(
                "Validation",
                "reward_on_geometry_not_changed",
                "Reward can only be set if end_episode_on_geometry_not_changed"
                " is true.",
            )
        if values["end_episode_on_geometry_not_changed"] and value is None:
            get_parser_logger().warning(
                "Please set a reward value for geometry not changed if episode"
                " should end on it. Will set 0 for you now, but this might not"
                " be you intention."
            )
            value = 0.0
        return value

    @validator("reward_on_episode_exceeds_max_timesteps", always=True)
    @classmethod
    def check_if_reward_given_if_max_steps_killer_activated(
        cls, value: float, values: Dict[str, Any]
    ) -> float:
        """Validator for reward_on_episode_exceeds_max_timesteps.

        Checks that 1) if a reward is set, also that the value for the
        max_timesteps_in_episode is greater than 0.
        2) If max_timesteps_in_episode is greater 0 a reward value is set.

        Args:
            value (float): value to validate
            values (Dict[str, Any]): previously validated values
            (here max_timesteps_in_episode is important)

        Raises:
            ParserException: Error is thrown if one of the conditions is
            not met.

        Returns:
            float: reward for the specified occurrence.
        """
        # due to pydantic this first statement should never be able to happen
        if "max_timesteps_in_episode" not in values:  # pragma: no cover
            raise ParserException(
                "Validation",
                "reward_on_episode_exceeds_max_timesteps",
                "Could not find definition of parameter "
                "max_timesteps_in_episode, please define this variable since "
                "otherwise this variable would have no function.",
            )
        if value is not None and (
            values["max_timesteps_in_episode"] is None
            or not values["max_timesteps_in_episode"]
        ):
            raise ParserException(
                "Validation",
                "reward_on_episode_exceeds_max_timesteps",
                "Reward can only be set if max_timesteps_in_episode is a "
                "positive integer.",
            )
        if values["max_timesteps_in_episode"] and value is None:
            get_parser_logger().warning(
                "Please set a reward value for max time steps exceeded, if "
                "episode should end on it. Will set 0 for you now, but this "
                "might not be you intention."
            )
            value = 0.0
        return value

    def should_add_callback(self) -> bool:
        """Bool function whether or not a validation callback is needed.

        Returns:
            bool: Return True if a validation callback is needed else False.
        """
        if self.validation_freq > 0:
            return True
        return False

    def get_callback(
        self,
        eval_environment: GymEnv,
        save_location: Optional[pathlib.Path] = None,
        normalizer_divisor: int = 1,
    ) -> EvalCallback:
        """Creates the EvalCallback with the values given in this object.

        Args:
            eval_environment (GymEnv): Evaluation environment. Should be the
                same as the normal training environment only that here the goal
                values should be set and not random.
            save_location (Optional[pathlib.Path]): Path to where the best
                models should be save to.
            normalizer_divisor (int, optional): Divisor for the eval_freq.
                Defaults to 1.

        Returns:
            EvalCallback: Validation callback parametrized by this object.
        """
        # build the dictionary holding the variables needed to initialize the
        # callback. See callback signature for variable explanation
        variable_dict = {
            "eval_env": eval_environment,
            "n_eval_episodes": len(self.validation_values),
            "eval_freq": int(self.validation_freq / normalizer_divisor),
        }
        # add variables for the case if the best model is to be saved
        if self.save_best_agent and save_location:
            variable_dict["best_model_save_path"] = (
                save_location / "eval/best_model"
            )
            variable_dict["log_path"] = save_location / "eval/log"
        return EvalCallback(**variable_dict)

    def end_validation(
        self, agent: BaseAlgorithm, environment: GymEnv
    ) -> Tuple[float, float]:
        """Function is called at the end of a validation.

        All clean up and last evaluation is going in here.

        Args:
            agent (BaseAlgorithm): Agent which is to be used to validate.
            environment (GymEnv): Validation environment

        Returns:
            Tuple[float, float]: See 'funct'evaluate_policy() for definition
        """
        variable_dict = {
            "model": agent,
            "env": environment,
            "n_eval_episodes": len(self.validation_values),
            "deterministic": True,
            "return_episode_rewards": True,
        }
        return evaluate_policy(**variable_dict)

    def get_environment_validation_parameters(self) -> Dict[str, Any]:
        """Gather the validation arguments used to initialize validator.

        Gets the validation parameters that need to be send to the environment
        if it gets converted to be a validation environment.

        Returns:
            Dict[str, Any]:
                dict with all the necessary parameters. Should mirror the
                parameters in parser_environment.Environment.set_validation
        """
        parameters_to_include = [
            "validation_values",
            "end_episode_on_geometry_not_changed",
            "max_timesteps_in_episode",
            "reward_on_episode_exceeds_max_timesteps",
            "reward_on_geometry_not_changed",
        ]
        return {key: getattr(self, key) for key in parameters_to_include}



from gymnasium import spaces

from releso.gym_environment import GymEnvironment


def test_gym_environment_dummy():
    # This test is just a dummy test since this class does not really do
    # anything since its functionality is provided by the parser environment.
    env = GymEnvironment(spaces.Discrete(10), spaces.Box(0, 1, shape=(1, 1)))
    env.step(1)
    env.reset()
    env.render()
    env.close()



"""Types and Type Unions defined for the ReLeSO package.

This file contains type definitions of commonly used items in this
framework/package. Mostly RL based here. Some more types can be found in other
files. Here only types with strictly external contents are created.
"""

from __future__ import annotations

from typing import Dict, Tuple, Union

from numpy import ndarray

from releso.util.module_import_raiser import ModuleImportRaiser

try:
    from gustaf import Faces, Volumes
except ModuleNotFoundError as err:
    Volumes = ModuleImportRaiser("gustaf", str(err))
    Faces = ModuleImportRaiser("gustaf", str(err))


#: Type definition for Observations.
ObservationType = Union[ndarray, Dict[str, ndarray], Tuple[ndarray, ...]]
#: Type definition for the info object returned for every step in the RL
#: Environment
InfoType = Dict
#: Type definition of the reward for every step in the RL Environment
RewardType = float
#: Type definition of the done marker for every step in the RL Environment
DoneType = bool

#: All possible classes that can make up a mesh
GustafMeshTypes = Union[Volumes, Faces]
#: Type definition for the return value of a Step. A step return the steps
#: observation, generated reward, whether or not the current step is complete
#: and additional info.
StepReturnType = Tuple[ObservationType, RewardType, DoneType, InfoType]



from typing import Any, Dict, List, Union
import pathlib

import pytest

from releso.base_parser import BaseParser
from releso.exceptions import ValidationNotSet


def recursive_remove_save_location(
    dict_to_clean: Union[List[Any], Dict[str, Any]]
):
    if isinstance(dict_to_clean, list):
        for item in dict_to_clean:
            recursive_remove_save_location(item)
    if isinstance(dict_to_clean, dict):
        if "save_location" in dict_to_clean:
            del dict_to_clean["save_location"]
        for value in dict_to_clean.values():
            if isinstance(value, dict):
                recursive_remove_save_location(value)
            elif isinstance(value, list):
                recursive_remove_save_location(value)


@pytest.mark.parametrize(
    [
        "verbosity",
        "basic_agent_definition",
        "environment",
        "number_of_timesteps",
        "number_of_episodes",
        "validation",
        "n_environments",
        "normalize_training_values",
        "multi_env_sequential",
    ],
    [
        (
            None,
            "PPO",
            None,
            10,
            None,
            None,
            None,
            None,
            None,
        ),
        (
            None,
            "PPO",
            None,
            10,
            None,
            {
                "validation_values": [120],
                "validation_freq": 1,
                "save_best_agent": False,
                "validate_on_training_end": False,
            },
            None,
            None,
            None,
        ),
    ],
    indirect=["basic_agent_definition"],
)
def test_base_parser_validation(
    verbosity,
    basic_agent_definition,
    environment,
    number_of_timesteps,
    number_of_episodes,
    validation,
    n_environments,
    normalize_training_values,
    multi_env_sequential,
    dir_save_location,
    basic_environment_definition,
    basic_verbosity_definition,
    clean_up_provider,
):
    calling_dict = {
        "save_location": dir_save_location,
        "verbosity": basic_verbosity_definition,
        "agent": basic_agent_definition,
        "environment": basic_environment_definition,
        "number_of_timesteps": number_of_timesteps,
    }
    if verbosity is not None:
        calling_dict["verbosity"].update(verbosity)
    if environment is not None:
        calling_dict["environment"].update(environment)
    if number_of_episodes is not None:
        calling_dict["number_of_episodes"] = number_of_episodes
    if validation is not None:
        calling_dict["validation"] = validation
    if n_environments is not None:
        calling_dict["n_environments"] = n_environments
    if normalize_training_values is not None:
        calling_dict["normalize_training_values"] = normalize_training_values
    if multi_env_sequential is not None:
        calling_dict["multi_env_sequential"] = multi_env_sequential

    # recursive_remove_save_location(calling_dict)
    # assert "save_location" not in calling_dict["environment"]
    # if this does not fail, the save_location is recursively added as it
    # should be
    base_parser = BaseParser(**calling_dict)

    file_paths = []
    with pytest.raises(RuntimeError) as err:
        file_paths.append(base_parser.save_model())
    assert "Please train the agent first" in str(err.value)
    if validation is None:
        with pytest.raises(ValidationNotSet) as err:
            base_parser.evaluate_model(throw_error_if_none=True)

        with pytest.raises(ValidationNotSet) as err:
            base_parser._create_validation_environment(
                throw_error_if_none=True
            )

        clean_up_provider(dir_save_location)
        return
    base_parser.evaluate_model()

    file_paths.extend(
        (base_parser.save_model(), base_parser.save_model("testing"))
    )

    for file_path in file_paths:
        print(file_path)
        file_path = pathlib.Path(file_path)
        assert file_path.exists()
        clean_up_provider(
            file_path.parent
        )  # be very careful here that it does not delete more than you want

    clean_up_provider(dir_save_location)


@pytest.mark.parametrize(
    [
        "verbosity",
        "basic_agent_definition",
        "agent_additions",
        "environment",
        "number_of_timesteps",
        "number_of_episodes",
        "validation",
        "n_environments",
        "normalize_training_values",
        "multi_env_sequential",
        "export_step_log",
    ],
    [
        (
            None,
            "PPO",
            {
                "n_steps": 64,
            },
            None,
            10,
            None,
            None,
            5,
            None,
            False,
            False,
        ),
        (
            None,
            "PPO",
            {
                "n_steps": 64,
            },
            None,
            10,
            5,
            None,
            5,
            True,
            None,
            False,
        ),
        (
            None,
            "PPO",
            {
                "n_steps": 64,
            },
            None,
            10,
            None,
            None,
            5,
            None,
            True,
            True,
        ),
        (
            None,
            "PPO",
            {
                "n_steps": 64,
            },
            None,
            10,
            None,
            None,
            5,
            True,
            False,
            False,
        ),
        (
            None,
            "PPO",
            {
                "n_steps": 64,
            },
            None,
            10,
            None,
            {
                "validation_values": [120],
                "validation_freq": 1,
                "save_best_agent": False,
                "validate_on_training_end": False,
            },
            None,
            None,
            None,
            False,
        ),
    ],
    indirect=["basic_agent_definition"],
)
def test_base_parser_learn(
    verbosity,
    basic_agent_definition,
    agent_additions,
    environment,
    number_of_timesteps,
    number_of_episodes,
    validation,
    n_environments,
    normalize_training_values,
    multi_env_sequential,
    export_step_log,
    dir_save_location,
    basic_environment_definition,
    basic_verbosity_definition,
    clean_up_provider,
):
    calling_dict = {
        "save_location": dir_save_location,
        "verbosity": basic_verbosity_definition,
        "agent": basic_agent_definition,
        "environment": basic_environment_definition,
        "number_of_timesteps": number_of_timesteps,
        "export_step_log": export_step_log,
    }
    if agent_additions is not None:
        calling_dict["agent"].update(agent_additions)
    if verbosity is not None:
        calling_dict["verbosity"].update(verbosity)
    if environment is not None:
        calling_dict["environment"].update(environment)
    if number_of_episodes is not None:
        calling_dict["number_of_episodes"] = number_of_episodes
    if validation is not None:
        calling_dict["validation"] = validation
    if n_environments is not None:
        calling_dict["n_environments"] = n_environments
    if normalize_training_values is not None:
        calling_dict["normalize_training_values"] = normalize_training_values
    if multi_env_sequential is not None:
        calling_dict["multi_env_sequential"] = multi_env_sequential

    # recursive_remove_save_location(calling_dict)
    # assert "save_location" not in calling_dict["environment"]
    # if this does not fail, the save_location is recursively added as it
    # should be
    base_parser = BaseParser(**calling_dict)
    base_parser.learn()
    clean_up_provider(dir_save_location)



"""Definition of the basic shape classes of releso.

In ReLeSO shapes mostly define what the actions in the RL problem are.
Due to that this framework is intended for shape optimization tasks the name
shape was chosen.

Note:
    To perform Free Form Deformation (FFD) splines are needed. These are given
    in a separate module, please see the file `spline.py` for these shape
    definitions.
"""
from collections import OrderedDict
from typing import Any, Dict, List, Optional

import numpy as np
from pydantic.class_validators import root_validator, validator
from pydantic.fields import PrivateAttr
from pydantic.types import conint

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.util.logger import get_parser_logger


class VariableLocation(BaseModel):
    """Variable location class.

    Object of this class defines the position and movement possibilities for a
    single dimensions of a single control point of the geometry.
    """

    #: coordinate of the current position of the current control point in the
    #: dimension
    current_position: float
    #: lower bound of possible values which can be reached by this variable
    min_value: Optional[float] = None
    #: upper bound  of possible values which can be reached by this variable
    max_value: Optional[float] = None
    #: number of steps the value range is divided into used to define the step
    #: if not given.
    n_steps: Optional[conint(ge=1)] = 10
    #: If discrete actions are used the step is used to define the new current
    #: position by adding/subtracting it from the current position.
    step: Optional[float] = None

    # non json variables
    #: Is true if min_value and max_value are not the same value
    _is_action: Optional[bool] = PrivateAttr(default=None)
    #: Original position needs to be saved so that the shape can be easily
    #: reset to its original state.
    _original_position: Optional[float] = PrivateAttr(default=None)

    def __init__(self, **data: Any) -> None:
        """Constructor for VariableLocation."""
        super().__init__(**data)
        self._original_position = self.current_position

    @validator("min_value", "max_value", always=True)
    @classmethod
    def set_variable_to_current_position_if_not_given(
        cls, v, values, field
    ) -> float:
        """Validator for min_value and max_value.

        Validation of the min and max values for the current VariableLocation.
        If non are set no variability is assumed min = max = current_position

        Args:
            v ([type]): Value to validate.
            values ([type]): Already validated values.
            field ([type]): Name of the field that is currently validated.

        Raises:
            ParserException: Parser error if current_position is not already
            validated.

        Returns:
            float: value of the validated value.
        """
        if v is None:
            if "current_position" in values.keys():
                return values["current_position"]
            else:
                raise ParserException(
                    "VariableLocation",
                    field,
                    "Please correctly define the current position.",
                )
        return v

    @validator("max_value", always=True)
    @classmethod
    def check_max_value_is_greater_than_current_value(
        cls, v, values, field
    ) -> float:
        """Validates that the max value is greater-equal to the current value.

        Args:
            v ([type]): Value to validate.
            values ([type]): Already validated values.
            field ([type]): Name of the field that is currently validated.

        Returns:
            float: value of the validated value.
        """
        if v < values["current_position"]:
            raise ParserException(
                "VariableLocation",
                field,
                f"The current_value {values['current_position']} must be"
                f" smaller or equal to the max_value {v}.",
            )
        return v

    @validator("min_value", always=True)
    @classmethod
    def check_min_value_is_smaller_than_current_value(
        cls, v, values, field
    ) -> float:
        """Validates that the min value is smaller-equal to the current value.

        Args:
            v ([type]): Value to validate.
            values ([type]): Already validated values.
            field ([type]): Name of the field that is currently validated.

        Returns:
            float: value of the validated value.
        """
        if v > values["current_position"]:
            raise ParserException(
                "VariableLocation",
                field,
                f"The current_value {values['current_position']} must be "
                f"bigger or equal to the max_value {v}.",
            )
        return v

    @root_validator
    @classmethod
    def define_step(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validator for class.

        Validation function that defines the step taken if the action would be
        discrete. If nothing is given for the calculation a n_steps default
        value of 10 is used.

        Args:
            values (Dict[str, Any]): Validated variables of the object.

        Returns:
            Dict[str, Any]: Validated variables but if steps was not given
            before now it has a value.
        """
        step, n_steps, min_value, max_value = (
            values.get("step"),
            values.get("n_steps"),
            values.get("min_value"),
            values.get("max_value"),
        )
        value_range = max_value - min_value
        if step is not None:
            if step > value_range:
                get_parser_logger().warning(
                    f"The defined step {step} is greater than the interval for"
                    f" this variable {value_range}. This is not intended "
                    "behavior."
                )
            return values  # if step length is defined ignore n_steps
        elif n_steps is None:
            # if discrete but neither step nor n_steps is defined a default 10
            # steps are assumed.
            n_steps = 10
        step = value_range / n_steps
        values["step"] = step
        return values

    def is_action(self) -> bool:
        """Boolean if variable defines an action.

        Checks if the variable has variability. Meaning is there optimization
        between the min_value and the max_value.

        Returns:
            bool: True if (self.max_value > self.current_position) or
            (self.min_value < self.current_position)
        """
        if self._is_action is None:
            self._is_action = (self.max_value > self.current_position) or (
                self.min_value < self.current_position
            )
        return self._is_action

    def apply_discrete_action(self, increasing: bool) -> float:
        """Apply a discrete action to the variable.

        Apply the a discrete action to the current value. Also applies clipping
        if min/max value is surpassed.

        Args:
            increasing (bool): Toggle whether or not the value should increase
            or not.

        Returns:
            float: Value of the position that the current position is now at.
        """
        step = self.step if increasing else -self.step
        self.current_position = np.clip(
            self.current_position + step, self.min_value, self.max_value
        )
        return self.current_position

    def apply_continuous_action(self, value: float) -> float:
        """Apply a continuous action the variable.

        Apply the zero-mean normalized value as the new current position.
        Needs to descale the value first. Also applies clipping.

        Args:
            value (float): Scaled value [-1,1] that needs to be descaled and
            applied as the new current position.

        Returns:
            float: New clipped value.
        """
        delta = self.max_value - self.min_value
        descaled_value = ((value + 1.0) / 2.0) * delta
        self.current_position = np.clip(
            descaled_value + self.min_value, self.min_value, self.max_value
        )
        return self.current_position

    def reset(self) -> None:
        """Resets the current_location to the initial position."""
        if self._original_position is not None:
            self.current_position = self._original_position


class ShapeDefinition(BaseModel):
    """Base of shape parameterization, also represents a simple point cloud."""

    #: control_points of the shape. These are the base variables used for the
    #: optimization. Overwrite `get_actions` and `get_parameter_values` if
    #: additional optimization variables are needed. See (WIP) NURBSDefinition.
    control_points: List[List[VariableLocation]]

    def get_number_of_points(self) -> int:
        """Returns the number of points in the Cube.

        Number of control points multiplied by the number of dimensions for
        each cp. Assumes that all dimensions have the same number of control
        points.

        Returns:
            int: number of points in the geometry
        """
        return int(len(self.control_points) * len(self.control_points[0]))

    @validator("control_points", each_item=True, pre=True)
    @classmethod
    def convert_all_control_point_locations_to_variable_locations(
        cls, v, values
    ):
        """Validator control_points.

        Converts all control points values into VariableLocations if the value
        is given as a simple float. Simple float will be converted into
        VariableLocation with current_position=value and no variability.

        Args:
            v ([type]): value to validate
            values ([type]): already validated values

        Returns:
            [type]: validated value
        """
        new_list = []
        for element in v:
            if type(element) is VariableLocation:
                new_list.append(element)
            elif type(element) in [dict, OrderedDict]:
                new_list.append(VariableLocation(**element))
            else:
                try:
                    element = float(element)
                    new_list.append(
                        VariableLocation(
                            current_position=element,
                            save_location=values["save_location"],
                        )
                    )
                except (ValueError, TypeError):
                    raise ParserException(
                        "ShapeDefinition",
                        "control_points",
                        "The control_points need to be either a float castable"
                        f" or a VariableLocation. Is of type {type(element)}.",
                    ) from None
        return new_list

    def get_parameter_values(self) -> List[List[float]]:
        """Returns the current positions of all control points.

        Returns:
            List[List[float]]: Positions of all control points.
        """
        return [
            [control_point.current_position for control_point in sub_list]
            for sub_list in self.control_points
        ]

    def get_actions(self) -> List[VariableLocation]:
        """Returns the action defined.

        Returns list of VariableLocations but only if the variable location is
        actually variable.

        Returns:
            List[VariableLocation]: See above
        """
        self.get_logger().debug(
            "Collecting all actions from shape parameterization."
        )
        return [
            variable
            for sub_dim in self.control_points
            for variable in sub_dim
            if variable.is_action()
        ]

    def get_shape(self) -> Any:
        """Generates the current shape.

        Here the control points are returned since there is no other meaning to
        this shape.

        Returns:
            Any: Shape that is generated.
        """
        return self.get_parameter_values()

    def reset(self) -> None:
        """Resets the shape to the original values."""
        for cp in self.control_points:
            for variable in cp:
                variable.reset()

    def draw_action_space(
        self,
        save_location: Optional[str] = None,
        no_axis: bool = False,
        fig_size: List[float] = None,
        dpi: int = 400,
    ):  # pragma: no cover
        """Draw the action space of the defined shape as a matplotlib figure.

        Needs to be reimplemented in subclasses where non control point
        variables are also used as actions. And if the control points are
        have more than 2D. The figure will plotted with matplotlib.

        Args:
            save_location (Optional[str], optional):
                Location to save the resulting figure to. Defaults to None.
            no_axis (bool, optional):
                Remove axis from the resulting figure. Defaults to False.
            fig_size (List[float], optional):
                Size of the resulting figure. Defaults to [6, 6].
            dpi (int, optional):
                DPI of the resulting figure. Defaults to 400.

        Raises:
            RuntimeError:
                Error is thrown if the control point's dimensionality is to
                high.
        """
        if fig_size is None:
            fig_size = [6, 6]
        import matplotlib.pyplot as plt
        from matplotlib.patches import Polygon

        control_points = self.control_points
        if len(self.control_points[0]) > 2:
            raise RuntimeError(
                "Could not draw the shapes action space. Only a 2D parametric"
                " space is currently available."
            )
        elif len(self.control_points[0]) == 1:
            for idx, cp in enumerate(control_points):
                cp.insert(
                    0,
                    VariableLocation(
                        current_location=float(idx),
                        save_location=self.save_location,
                    ),
                )
        phi = np.linspace(0, 2 * np.pi, len(control_points))
        rgb_cycle = np.vstack(
            (  # Three sinusoids
                0.5 * (1.0 + np.cos(phi)),  # scaled to [0,1]
                0.5
                * (1.0 + np.cos(phi + 2 * np.pi / 3)),  # 120° phase shifted.
                0.5 * (1.0 + np.cos(phi - 2 * np.pi / 3)),
            )
        ).T  # Shape = (60,3)
        fig, ax = plt.subplots(figsize=fig_size, dpi=dpi)

        dots = [[], []]

        for elem, color in zip(control_points, rgb_cycle):
            dots[0].append(elem[0].current_position)
            dots[1].append(elem[1].current_position)
            cur_pos = np.array([dots[0][-1], dots[1][-1]])
            spanning_elements = []
            no_boundary = False
            for item in elem:
                spanning_elements.append([item.min_value, item.max_value])
                if item.min_value == item.max_value:
                    no_boundary = True
            boundary = []
            for i, j in zip([0, 1, 1, 0], [0, 0, 1, 1]):
                end_pos = np.array(
                    [spanning_elements[0][i], spanning_elements[1][j]]
                )
                if not np.isclose(cur_pos, end_pos).all():  # draw arrow
                    difference = end_pos - cur_pos
                    ax.arrow(
                        cur_pos[0],
                        cur_pos[1],
                        difference[0] * 0.9,
                        difference[1] * 0.9,
                        width=0.005,
                        color=color,
                    )
                boundary.append(end_pos)
            if not no_boundary:
                pol = Polygon(
                    boundary, facecolor=color, linewidth=1, alpha=0.2
                )
                ax.add_patch(pol)
        ax.scatter(dots[0], dots[1], c=rgb_cycle, marker="o", s=50, zorder=3)
        if no_axis:
            plt.axis("off")
        if save_location:
            fig.savefig(save_location, transparent=True)
            plt.close()



import copy
import pathlib

import gymnasium as gym
import numpy as np
import pytest
from conftest import dir_save_location_path
from pydantic import ValidationError

from releso.parser_environment import Environment, MultiProcessing
from releso.verbosity import VerbosityLevel


@pytest.mark.parametrize(
    ["n_cores", "w_cores", "error_n_cores"],
    [(1, 1, False), (False, 1, False), (12, 12, False), (-10, False, True)],
)
def test_parser_environment_multiprocessing(
    n_cores, w_cores, error_n_cores, dir_save_location
):
    calling_dict = {
        "save_location": dir_save_location,
    }
    if n_cores:
        calling_dict["number_of_cores"] = n_cores
    if error_n_cores:
        with pytest.raises(ValidationError):
            MultiProcessing(**calling_dict)
        return

    multi = MultiProcessing(**calling_dict)

    assert multi.number_of_cores == w_cores


def test_parser_environment_environment(dir_save_location, default_shape):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    env.close()


@pytest.mark.parametrize(
    [
        "n_cores",
        "w_cores",
        "max_timesteps_in_episode",
        "end_episode_on_geometry_not_changed",
        "reward_on_geometry_not_changed",
        "reward_on_episodes_exceeds_max_timesteps",
        "error",
        "warning",
    ],
    [
        (None, 1, None, None, None, None, False, False),
        (
            None,
            1,
            None,
            None,
            None,
            -1,
            "Reward can only be set if max_timesteps_in_episode a positive",
            False,
        ),
        # (
        #     None,
        #     1,
        #     False,
        #     None,
        #     None,
        #     -1,
        #     "Reward can only be set if max_timesteps_in_episode a positive",
        #     False,
        # ), # Somewhat strange behavior, therefore not used as test
        (
            None,
            1,
            10,
            None,
            None,
            None,
            False,
            "Please set a reward value for max time steps exceeded, if",
        ),
        (12, 12, 10, None, None, 0.1, False, False),
        (
            2,
            2,
            None,
            True,
            None,
            None,
            False,
            "Please set a reward value for geometry not changed",
        ),
        (
            None,
            None,
            None,
            None,
            1,
            None,
            "Reward can only be set if end_episode_on_geometry_not_changed ",
            False,
        ),
        (
            None,
            None,
            None,
            False,
            1,
            None,
            "Reward can only be set if end_episode_on_geometry_not_changed ",
            False,
        ),
    ],
)
def test_parser_environment_test_parsing(
    n_cores,
    w_cores,
    max_timesteps_in_episode,
    end_episode_on_geometry_not_changed,
    reward_on_geometry_not_changed,
    reward_on_episodes_exceeds_max_timesteps,
    error,
    warning,
    dir_save_location,
    caplog,
    default_shape,
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    if n_cores is not None:
        calling_dict["multi_processing"]["number_of_cores"] = n_cores
    if max_timesteps_in_episode is not None:
        calling_dict["max_timesteps_in_episode"] = max_timesteps_in_episode
    if end_episode_on_geometry_not_changed is not None:
        calling_dict[
            "end_episode_on_geometry_not_changed"
        ] = end_episode_on_geometry_not_changed
    if reward_on_geometry_not_changed is not None:
        calling_dict[
            "reward_on_geometry_not_changed"
        ] = reward_on_geometry_not_changed
    if reward_on_episodes_exceeds_max_timesteps is not None:
        calling_dict[
            "reward_on_episode_exceeds_max_timesteps"
        ] = reward_on_episodes_exceeds_max_timesteps
    if error:
        with pytest.raises(ValidationError) as err:
            Environment(**calling_dict)
        assert error in str(err.value)
        return
    with caplog.at_level("WARNING"):
        env = Environment(**calling_dict)
    if warning:
        assert warning in caplog.text
    assert env.is_multiprocessing() == w_cores


def test_parser_environment_observations_no_observation(
    dir_save_location, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": False,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    with pytest.raises(RuntimeError) as err:
        env._define_observation_space()
    assert "The observation space is empty." in str(err.value)


def test_parser_environment_observations_single(
    dir_save_location, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    obs_space = env._define_observation_space()
    assert obs_space.shape == (2,)
    observations = {}
    observations["geometry_observation"] = env.geometry.get_observation()
    for step in env.spor.steps:
        observations.update(step.get_default_observation(observations))
    new_obs = np.array(env.check_observations(observations))
    assert new_obs.shape == (2,)


def test_parser_environment_observations_compress(
    dir_save_location, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                    "additional_observations": [
                        {"name": "test1", "value_min": 0.0, "value_max": 1.0},
                        {"name": "test2", "value_min": 0.0, "value_max": 1.0},
                        {
                            "name": "test3",
                            "value_min": 0.0,
                            "value_max": 1.0,
                            "observation_shape": [
                                9,
                            ],
                            "value_type": "float",
                        },
                    ],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    obs_space = env._define_observation_space()
    assert env._flatten_observations
    assert obs_space.shape == (13,)
    observations = {}
    observations["geometry_observation"] = env.geometry.get_observation()
    for step in env.spor.steps:
        observations.update(step.get_default_observation(observations))
    new_obs = np.array(env.check_observations(observations))
    assert new_obs.shape == (13,)


def test_parser_environment_observations_cnn(dir_save_location, default_shape):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "name": "internal",
                    "save_location": dir_save_location,
                    "reward_on_error": 0.0,
                    "use_communication_interface": True,
                    "working_directory": str(dir_save_location),
                    "function_name": "xns_cnn",
                },
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    obs_space = env._define_observation_space()
    assert not env._flatten_observations
    n_envs = 0
    for k in obs_space.spaces.keys():
        assert k in ["cnn_observation", "geometry_observation"]
        n_envs += 1
    assert n_envs == 2
    observations = {}
    observations["geometry_observation"] = env.geometry.get_observation()
    for step in env.spor.steps:
        observations.update(step.get_default_observation(observations))
    new_obs = env.check_observations(observations)
    assert isinstance(new_obs, dict)


def test_parser_environment_observations_non_compressible(
    dir_save_location, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                    "additional_observations": [
                        {"name": "test1", "value_min": 0.0, "value_max": 1.0},
                        {"name": "test2", "value_min": 0.0, "value_max": 1.0},
                        {
                            "name": "test3",
                            "value_min": 0.0,
                            "value_max": 1.0,
                            "observation_shape": [
                                9,
                                2,
                            ],
                            "value_type": "float",
                        },
                    ],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    obs_space = env._define_observation_space()
    assert not env._flatten_observations
    n_envs = 0
    for k in obs_space.spaces.keys():
        assert k in ["test1", "test2", "test3", "geometry_observation"]
        n_envs += 1
    assert n_envs == 4


def test_parser_environment_observations_step_reset_simple(
    dir_save_location, clean_up_provider, caplog, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)

    gym_env = env.get_gym_environment(
        {
            "logger_name": "test",
            "log_file_location": dir_save_location,
            "logging_level": VerbosityLevel.INFO,
        }
    )
    # clean_up_provider(dir_save_location)
    assert issubclass(type(gym_env), gym.Env)

    obs, info = gym_env.reset()
    assert type(obs) in [list, np.ndarray]
    assert isinstance(info, dict)
    assert len(obs) == 2

    obs, _, _, _, info = gym_env.step(1)
    assert pytest.approx(obs) == [4.9, 5.0]

    with caplog.at_level(VerbosityLevel.WARNING):
        _ = env.get_gym_environment()

    clean_up_provider(dir_save_location)


def test_parser_environment_max_timesteps(
    dir_save_location, clean_up_provider, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
        "max_timesteps_in_episode": 5,
        "reward_on_episode_exceeds_max_timesteps": -1,
    }
    env = Environment(**calling_dict)

    gym_env = env.get_gym_environment()
    gym_env.reset()
    for idx in range(4):
        obs, _, done, _, info = gym_env.step(1)
        assert pytest.approx(obs) == [5 - ((idx + 1) * 0.1), 5.0]
        assert not done
    obs, _, done, _, info = gym_env.step(1)
    assert pytest.approx(obs) == [4.5, 5.0]
    assert done
    assert info["reset_reason"] == "max_timesteps_exceeded"

    clean_up_provider(dir_save_location)


def test_parser_environment_geometry_not_changed(
    dir_save_location, clean_up_provider, default_shape
):
    dir_save_location_path(True)
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": True,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "execution_command": "echo",
                    "command_options": ["'test'"],
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
        "end_episode_on_geometry_not_changed": True,
        "reward_on_geometry_not_changed": -1,
    }
    env = Environment(**calling_dict)

    gym_env = env.get_gym_environment()
    gym_env.reset()
    for idx in range(4):
        obs, _, done, _, info = gym_env.step(1)
        assert pytest.approx(obs) == [5 - ((idx + 1) * 0.1), 5.0]
        assert not done
    # this should increase the second variable which is at its bound so it will
    #  not change
    obs, _, done, _, info = gym_env.step(2)
    assert pytest.approx(obs) == [4.6, 5.0]
    assert done
    assert info["reset_reason"] == "geometry_not_changed"

    clean_up_provider(dir_save_location)


def test_parser_environment_non_geometry_observations(
    dir_save_location, clean_up_provider, default_shape
):
    dir_save_location_path(True)
    file_path = pathlib.Path(__file__).parent
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": False,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "python_file_path": (
                        f"{str(file_path)}/samples"
                        "/spor_python_scripts_tests/file_exists_has_main.py"
                    ),
                    "use_communication_interface": True,
                    "additional_observations": 3,
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)

    gym_env = env.get_gym_environment()
    gym_env.reset()
    obs, _, _, _, _ = gym_env.step(2)
    assert obs == pytest.approx([1, 2, 3])
    clean_up_provider(dir_save_location)


def test_parser_environment_validation(
    dir_save_location, clean_up_provider, default_shape
):
    dir_save_location_path(True)
    file_path = pathlib.Path(__file__).parent
    calling_dict = {
        "save_location": dir_save_location,
        "geometry": {
            "save_location": dir_save_location,
            "shape_definition": default_shape,
            "action_based_observation": False,
        },
        "spor": {
            "save_location": dir_save_location,
            "steps": [
                {
                    "save_location": dir_save_location,
                    "name": "test",
                    "reward_on_error": -1,
                    "working_directory": str(dir_save_location),
                    "python_file_path": (
                        f"{str(file_path)}/samples"
                        "/spor_python_scripts_tests/file_exists_has_main.py"
                    ),
                    "use_communication_interface": True,
                    "additional_observations": 3,
                }
            ],
            "reward_aggregation": "sum",
        },
        "multi_processing": {
            "save_location": dir_save_location,
        },
    }
    env = Environment(**calling_dict)
    val_values = [1, 2, 3]
    env.set_validation(
        validation_values=copy.deepcopy(val_values),
    )
    gym_env = env.get_gym_environment()
    gym_env.reset()
    for _ in range(len(val_values)):
        _ = gym_env.step(2)
        assert (val_v := env.get_validation_id()) in val_values
        gym_env.reset()
        val_values.remove(val_v)
    assert len(val_values) == 0
    clean_up_provider(dir_save_location)
